import json
import mock
import re
import unittest

from kinto.core.testing import get_user_headers

from .support import (BaseWebTest, MINIMALIST_RECORD,
                      MINIMALIST_GROUP, MINIMALIST_BUCKET,
                      MINIMALIST_COLLECTION)


class RecordsViewTest(BaseWebTest, unittest.TestCase):

    collection_url = '/buckets/beers/collections/barley/records'
    _record_url = '/buckets/beers/collections/barley/records/{}'

    def setUp(self):
        super().setUp()
        self.app.put_json('/buckets/beers', MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_RECORD,
                                  headers=self.headers)
        self.record = resp.json['data']
        self.record_url = self._record_url.format(self.record['id'])

    def test_records_can_be_accessed_by_id(self):
        self.app.get(self.record_url, headers=self.headers)

    def test_unknown_bucket_raises_403(self):
        other_bucket = self.collection_url.replace('beers', 'sodas')
        self.app.get(other_bucket, headers=self.headers, status=403)

    def test_unknown_collection_raises_404(self):
        other_collection = self.collection_url.replace('barley', 'pills')
        resp = self.app.get(other_collection, headers=self.headers, status=404)
        self.assertEqual(resp.json['details']['id'], 'pills')
        self.assertEqual(resp.json['details']['resource_name'], 'collection')

    def test_unknown_record_raises_404(self):
        other_record = self.record_url.replace(self.record['id'], self.record['id']+'blah')
        response = self.app.get(other_record, headers=self.headers, status=404)
        self.assertEqual(response.json['details']['id'], self.record['id']+'blah')
        self.assertEqual(response.json['details']['resource_name'], 'record')

    def test_unknown_collection_does_not_query_timestamp(self):
        other_collection = self.collection_url.replace('barley', 'pills')
        patch = mock.patch.object(self.app.app.registry.storage,
                                  'collection_timestamp')
        self.addCleanup(patch.stop)
        mocked = patch.start()
        self.app.get(other_collection, headers=self.headers, status=404)
        self.assertFalse(mocked.called)

    def test_parent_collection_is_fetched_only_once_in_batch(self):
        batch = {'requests': []}
        nb_create = 25
        for i in range(nb_create):
            request = {'method': 'POST',
                       'path': self.collection_url,
                       'body': MINIMALIST_RECORD}
            batch['requests'].append(request)

        with mock.patch.object(self.storage, 'get',
                               wraps=self.storage.get) as patched:
            self.app.post_json('/batch', batch, headers=self.headers)
            self.assertEqual(patched.call_count, 1)

    def test_individual_collections_can_be_deleted(self):
        resp = self.app.get(self.collection_url, headers=self.headers)
        self.assertEqual(len(resp.json['data']), 1)
        self.app.delete(self.collection_url, headers=self.headers)
        resp = self.app.get(self.collection_url, headers=self.headers)
        self.assertEqual(len(resp.json['data']), 0)

    def test_records_can_be_added_to_collections(self):
        response = self.app.get(self.record_url, headers=self.headers)
        record = response.json['data']
        del record['id']
        del record['last_modified']
        self.assertEquals(record, MINIMALIST_RECORD['data'])

    def test_records_are_isolated_by_bucket_and_by_collection(self):
        # By collection.
        self.app.put_json('/buckets/beers/collections/pills',
                          MINIMALIST_BUCKET,
                          headers=self.headers)
        other_collection = self.record_url.replace('barley', 'pills')
        self.app.get(other_collection, headers=self.headers, status=404)

        # By bucket.
        self.app.put_json('/buckets/sodas',
                          MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.put_json('/buckets/sodas/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)
        other_bucket = self.record_url.replace('beers', 'sodas')
        self.app.get(other_bucket, headers=self.headers, status=404)

        # By bucket and by collection.
        self.app.put_json('/buckets/be',
                          MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.put_json('/buckets/be/collections/ba',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)
        other = self.record_url.replace('barley', 'ba').replace('beers', 'be')
        self.app.get(other, headers=self.headers, status=404)

    def test_a_collection_named_group_do_not_interfere_with_groups(self):
        # Create a group.
        self.app.put_json('/buckets/beers/groups/test',
                          MINIMALIST_GROUP,
                          headers=self.headers)
        # Create a record in a collection named "group".
        self.app.put_json('/buckets/beers/collections/groups',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)
        collection_group = self.collection_url.replace('barley', 'groups')
        self.app.post_json(collection_group,
                           MINIMALIST_RECORD,
                           headers=self.headers)
        # There is still only one group.
        resp = self.app.get('/buckets/beers/groups', headers=self.headers)
        self.assertEqual(len(resp.json['data']), 1)

    def test_records_can_be_filtered_on_any_field(self):
        self.app.post_json(self.collection_url,
                           MINIMALIST_RECORD,
                           headers=self.headers)
        response = self.app.get(self.collection_url + '?unknown=1',
                                headers=self.headers)
        self.assertEqual(len(response.json['data']), 0)

    def test_records_can_be_sorted_on_any_field(self):
        for i in range(3):
            record = {**MINIMALIST_RECORD, 'data': {
                **MINIMALIST_RECORD['data'],
                'name': 'Stout {}'.format(i)}
            }
            self.app.post_json(self.collection_url,
                               record,
                               headers=self.headers)

        response = self.app.get(self.collection_url + '?_sort=-name',
                                headers=self.headers)
        names = [i['name'] for i in response.json['data']]
        self.assertEqual(names,
                         ['Stout 2', 'Stout 1', 'Stout 0', 'Hulled Barley'])

    def test_wrong_create_permissions_cannot_be_added_on_records(self):
        record = {**MINIMALIST_RECORD, 'permissions': {'record:create': ['fxa:user']}}
        self.app.put_json(self.record_url,
                          record,
                          headers=self.headers,
                          status=400)

    def test_create_a_record_update_collection_timestamp(self):
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        old_timestamp = int(json.loads(collection_resp.headers['ETag']))
        self.app.post_json(self.collection_url,
                           MINIMALIST_RECORD,
                           headers=self.headers,
                           status=201)
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        new_timestamp = int(json.loads(collection_resp.headers['ETag']))
        assert old_timestamp < new_timestamp

    def test_create_a_record_without_id_generates_a_uuid(self):
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_RECORD,
                                  headers=self.headers,
                                  status=201)
        regexp = re.compile(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-'
                            r'[0-9a-f]{4}-[0-9a-f]{12}$')
        self.assertTrue(regexp.match(resp.json['data']['id']))

    def test_create_a_record_with_an_id_uses_it(self):
        record = {'data': dict(id='a-simple-id', **MINIMALIST_RECORD['data'])}
        resp = self.app.post_json(self.collection_url,
                                  record,
                                  headers=self.headers,
                                  status=201)
        self.assertEqual(resp.json['data']['id'], 'a-simple-id')

    def test_create_a_record_with_an_existing_id_returns_existing(self):
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_RECORD,
                                  headers=self.headers,
                                  status=201)
        existing_id = resp.json['data']['id']
        record = {'data': {'id': existing_id, 'stars': 8}}
        resp = self.app.post_json(self.collection_url,
                                  record,
                                  headers=self.headers,
                                  status=200)
        self.assertNotIn('stars', resp.json['data'])

    def test_create_a_record_with_existing_from_someone_else_gives_403(self):
        resp = self.app.post_json(self.collection_url,
                                  MINIMALIST_RECORD,
                                  headers=self.headers,
                                  status=201)
        existing_id = resp.json['data']['id']
        record = {'data': {'id': existing_id, 'stars': 8}}
        resp = self.app.post_json(self.collection_url,
                                  record,
                                  headers=get_user_headers('tartanpion'),
                                  status=403)

    def test_update_a_record_update_collection_timestamp(self):
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        old_timestamp = int(json.loads(collection_resp.headers['ETag']))
        self.app.put_json(self.record_url,
                          MINIMALIST_RECORD,
                          headers=self.headers,
                          status=200)
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        new_timestamp = int(json.loads(collection_resp.headers['ETag']))
        assert old_timestamp < new_timestamp

    def test_delete_a_record_update_collection_timestamp(self):
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        old_timestamp = int(json.loads(collection_resp.headers['ETag']))
        self.app.delete(self.record_url,
                        headers=self.headers,
                        status=200)
        collection_resp = self.app.get(self.collection_url,
                                       headers=self.headers)
        new_timestamp = int(json.loads(collection_resp.headers['ETag']))
        assert old_timestamp < new_timestamp

    def test_record_is_accessible_by_group_member(self):
        # access as aaron
        self.aaron_headers = {**self.headers, **get_user_headers('aaron')}

        resp = self.app.get('/',
                            headers=self.aaron_headers,
                            status=200)

        self.create_group('beers', 'brewers', [resp.json['user']['id']])
        record = {**MINIMALIST_RECORD, 'permissions': {'read': ['/buckets/beers/groups/brewers']}}
        self.app.put_json(self.record_url,
                          record,
                          headers=self.headers,
                          status=200)

        self.app.get(self.record_url,
                     headers=self.aaron_headers,
                     status=200)

    def test_records_should_reject_unaccepted_request_content_type(self):
        headers = {**self.headers, 'Content-Type': 'text/plain'}
        self.app.put(self.record_url,
                     MINIMALIST_RECORD,
                     headers=headers,
                     status=415)

    def test_records_should_reject_unaccepted_client_accept(self):
        headers = {**self.headers, 'Accept': 'text/plain'}
        self.app.get(self.record_url,
                     MINIMALIST_RECORD,
                     headers=headers,
                     status=406)

    def test_records_should_accept_client_accept(self):
        headers = {**self.headers, 'Accept': '*/*'}
        self.app.get(self.record_url,
                     MINIMALIST_RECORD,
                     headers=headers,
                     status=200)

    def test_records_can_be_created_after_deletion(self):
        self.app.delete(self.record_url,
                        headers=self.headers,
                        status=200)
        headers = {**self.headers, 'If-None-Match': '*'}
        self.app.put_json(self.record_url, MINIMALIST_RECORD,
                          headers=headers, status=201)


class RecordsViewMergeTest(BaseWebTest, unittest.TestCase):

    collection_url = '/buckets/beers/collections/barley/records'
    _record_url = '/buckets/beers/collections/barley/records/{}'

    def setUp(self):
        super().setUp()
        self.app.put_json('/buckets/beers', MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)
        record = {**MINIMALIST_RECORD, 'data': {'grain': {'one': 1}}}
        resp = self.app.post_json(self.collection_url, record,
                                  headers=self.headers)
        self.record = resp.json['data']
        self.record_url = self._record_url.format(self.record['id'])

    def test_merge_patch(self):
        headers = {**self.headers, 'Content-Type': 'application/merge-patch+json'}
        json = {'data': {'grain': {'two': 2}}}
        resp = self.app.patch_json(self.record_url,
                                   json,
                                   headers=headers,
                                   status=200)
        self.assertEquals(resp.json['data']['grain']['one'], 1)
        self.assertEquals(resp.json['data']['grain']['two'], 2)

    def test_merge_patch_remove_nones(self):
        headers = {**self.headers, 'Content-Type': 'application/merge-patch+json'}
        json = {'data': {'grain': {'one': None}}}
        resp = self.app.patch_json(self.record_url,
                                   json,
                                   headers=headers,
                                   status=200)
        self.assertNotIn('one', resp.json['data']['grain'])


class RecordsViewPatchTest(BaseWebTest, unittest.TestCase):

    collection_url = '/buckets/beers/collections/barley/records'
    _record_url = '/buckets/beers/collections/barley/records/{}'

    def setUp(self):
        super().setUp()
        self.patch_headers = {**self.headers, 'Content-Type': 'application/json-patch+json'}

        self.app.put_json('/buckets/beers', MINIMALIST_BUCKET,
                          headers=self.headers)
        self.app.put_json('/buckets/beers/collections/barley',
                          MINIMALIST_COLLECTION,
                          headers=self.headers)
        record = {**MINIMALIST_RECORD, 'permissions': {
            'read': ['alice', 'carla'],
            'write': ['bob']
        }}
        resp = self.app.post_json(self.collection_url,
                                  record,
                                  headers=self.headers)
        self.record = resp.json['data']
        self.record_url = self._record_url.format(self.record['id'])

    def test_patch_add_permissions(self):
        json = [{'op': 'add', 'path': '/permissions/read/me', 'value': 'me'}]
        resp = self.app.patch_json(self.record_url,
                                   json,
                                   headers=self.patch_headers,
                                   status=200)

        perms = resp.json['permissions']
        self.assertIn('me', perms['read'])
        self.assertIn('alice', perms['read'])

    def test_patch_update_permissions(self):
        json = [{'op': 'add', 'path': '/permissions/read', 'value': ['me']}]
        resp = self.app.patch_json(self.record_url,
                                   json,
                                   headers=self.patch_headers,
                                   status=200)

        perms = resp.json['permissions']
        self.assertIn('me', perms['read'])
        self.assertNotIn(('alice', 'bob'), perms['read'])

    def test_patch_remove_permissions(self):
        json = [{'op': 'remove', 'path': '/permissions/read/alice'}]
        resp = self.app.patch_json(self.record_url,
                                   json,
                                   headers=self.patch_headers,
                                   status=200)
        perms = resp.json['permissions']
        self.assertNotIn('alice', perms['read'])

    def test_patch_move_permissions(self):
        json = [
            {'op': 'move', 'from': '/permissions/read/alice',
                           'path': '/data/old'}
        ]
        resp = self.app.patch_json(self.record_url,
                                   json,
                                   headers=self.patch_headers,
                                   status=200)

        perms = resp.json['permissions']
        data = resp.json['data']
        self.assertNotIn('alice', perms['read'])
        self.assertEquals('alice', data['old'])

    def test_patch_raises_400_on_wrong_path(self):
        json = [{'op': 'add', 'path': '/permissions/destroy/me'}]
        self.app.patch_json(self.record_url,
                            json,
                            headers=self.patch_headers,
                            status=400)



import mock
import unittest

import transaction
import pytest
from pyramid import testing

from kinto import main as kinto_main
from kinto.core.errors import ERRORS
from kinto.core.storage.exceptions import RecordNotFoundError
from kinto.core.testing import FormattedErrorMixin, sqlalchemy, skip_if_no_statsd
from kinto.plugins.quotas.listener import (
    QUOTA_RESOURCE_NAME, QUOTA_BUCKET_ID, QUOTA_COLLECTION_ID)
from kinto.plugins.quotas.utils import record_size

from .. import support


class PluginSetup(unittest.TestCase):

    @skip_if_no_statsd
    def test_a_statsd_timer_is_used_for_quotas_if_configured(self):
        settings = {
            "statsd_url": "udp://127.0.0.1:8125",
            "includes": "kinto.plugins.quotas"
        }
        config = testing.setUp(settings=settings)
        with mock.patch('kinto.core.statsd.Client.timer') as mocked:
            kinto_main(None, config=config)
            mocked.assert_called_with('plugins.quotas')


class QuotaWebTest(support.BaseWebTest, unittest.TestCase):

    bucket_uri = '/buckets/test'
    collection_uri = '/buckets/test/collections/col'
    record_uri = '/buckets/test/collections/col/records/rec'
    group_uri = '/buckets/test/groups/grp'

    @classmethod
    def setUpClass(cls):
        if not sqlalchemy:
            raise unittest.SkipTest("postgresql is not installed.")

    def create_bucket(self):
        resp = self.app.put(self.bucket_uri, headers=self.headers)
        self.bucket = resp.json['data']

    def create_collection(self):
        resp = self.app.put(self.collection_uri, headers=self.headers)
        self.collection = resp.json['data']

    def create_group(self):
        body = {'data': {'members': ['elle']}}
        resp = self.app.put_json(self.group_uri, body, headers=self.headers)
        self.group = resp.json['data']

    def create_record(self):
        body = {'data': {'foo': 42}}
        resp = self.app.put_json(self.record_uri, body, headers=self.headers)
        self.record = resp.json['data']

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)

        # Setup the postgresql backend for transaction support.
        settings['storage_backend'] = 'kinto.core.storage.postgresql'
        db = "postgres://postgres:postgres@localhost/testdb"
        settings['storage_url'] = db
        settings['permission_backend'] = 'kinto.core.permission.postgresql'
        settings['permission_url'] = db
        settings['cache_backend'] = 'kinto.core.cache.memory'

        settings['includes'] = 'kinto.plugins.quotas'
        return settings

    def assertStatsEqual(self, data, stats):
        for key in stats:
            assert data[key] == stats[key]


class HelloViewTest(QuotaWebTest):

    def test_quota_capability_if_enabled(self):
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('quotas', capabilities)


class QuotaListenerTest(QuotaWebTest):

    #
    # Bucket
    #
    def test_quota_tracks_bucket_creation(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        storage_size += record_size(self.record)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 1,
            "storage_size": storage_size
        })

    def test_tracks_bucket_attributes_update(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        body = {'data': {'foo': 'baz'}}
        resp = self.app.patch_json(self.bucket_uri, body,
                                   headers=self.headers)
        storage_size = record_size(resp.json['data'])
        storage_size += record_size(self.collection)
        storage_size += record_size(self.record)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 1,
            "storage_size": storage_size
        })

    def test_bucket_delete_destroys_its_quota_entries(self):
        self.create_bucket()
        self.app.delete(self.bucket_uri, headers=self.headers)
        stored_in_backend, _ = self.storage.get_all(
            parent_id='/buckets/test',
            collection_id=QUOTA_RESOURCE_NAME)
        assert len(stored_in_backend) == 0

    def test_bucket_delete_doesnt_raise_if_quota_entries_do_not_exist(self):
        self.create_bucket()
        self.storage.delete(parent_id='/buckets/test',
                            collection_id=QUOTA_RESOURCE_NAME,
                            object_id=QUOTA_BUCKET_ID)
        transaction.commit()
        self.app.delete(self.bucket_uri, headers=self.headers)

    #
    # Collection
    #
    def test_stats_are_not_accessible_if_collection_does_not_exist(self):
        self.create_bucket()
        self.app.get(self.collection_uri, headers=self.headers, status=404)

    def test_quota_tracks_collection_creation(self):
        self.create_bucket()
        self.create_collection()

        # Bucket stats
        storage_size = record_size(self.bucket) + record_size(self.collection)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 0,
            "storage_size": storage_size
        })

        # Collection stats
        storage_size = record_size(self.collection)
        data = self.storage.get(QUOTA_RESOURCE_NAME, self.collection_uri,
                                QUOTA_COLLECTION_ID)
        self.assertStatsEqual(data, {
            "record_count": 0,
            "storage_size": storage_size
        })

    def test_tracks_collection_attributes_update(self):
        self.create_bucket()
        self.create_collection()
        body = {'data': {'foo': 'baz'}}
        resp = self.app.patch_json(self.collection_uri, body,
                                   headers=self.headers)
        # Bucket stats
        storage_size = record_size(self.bucket)
        storage_size += record_size(resp.json['data'])

        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 0,
            "storage_size": storage_size
        })

        # Collection stats
        storage_size -= record_size(self.bucket)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.collection_uri,
                                object_id=QUOTA_COLLECTION_ID)
        self.assertStatsEqual(data, {
            "record_count": 0,
            "storage_size": storage_size
        })

    def test_tracks_collection_delete(self):
        self.create_bucket()
        self.create_collection()
        body = {'data': {'foo': 'baz'}}
        self.app.patch_json(self.collection_uri, body,
                            headers=self.headers)
        self.app.delete(self.collection_uri, headers=self.headers)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 0,
            "record_count": 0,
            "storage_size": record_size(self.bucket)
        })

    def test_collection_delete_destroys_its_quota_entries(self):
        self.create_bucket()
        self.create_collection()
        self.app.delete(self.collection_uri, headers=self.headers)
        stored_in_backend, _ = self.storage.get_all(
            parent_id=self.collection_uri,
            collection_id=QUOTA_RESOURCE_NAME)
        assert len(stored_in_backend) == 0

    def test_collection_delete_doesnt_raise_if_quota_entries_dont_exist(self):
        self.create_bucket()
        self.create_collection()
        self.storage.delete(parent_id=self.collection_uri,
                            collection_id=QUOTA_RESOURCE_NAME,
                            object_id=QUOTA_COLLECTION_ID)
        transaction.commit()
        self.app.delete(self.collection_uri, headers=self.headers)

    def test_tracks_collection_delete_with_multiple_records(self):
        self.create_bucket()
        self.create_collection()
        body = {'data': {'foo': 42}}
        self.app.post_json('{}/records'.format(self.collection_uri),
                           body, headers=self.headers)
        self.app.post_json('{}/records'.format(self.collection_uri),
                           body, headers=self.headers)
        self.app.post_json('{}/records'.format(self.collection_uri),
                           body, headers=self.headers)
        self.app.post_json('{}/records'.format(self.collection_uri),
                           body, headers=self.headers)
        self.app.delete(self.collection_uri, headers=self.headers)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 0,
            "record_count": 0,
            "storage_size": record_size(self.bucket)
        })

    #
    # Group
    #

    def test_quota_tracks_group_creation(self):
        self.create_bucket()
        self.create_group()
        storage_size = record_size(self.bucket) + record_size(self.group)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 0,
            "record_count": 0,
            "storage_size": storage_size
        })

    def test_tracks_group_attributes_update(self):
        self.create_bucket()
        self.create_group()
        body = {'data': {'foo': 'baz', 'members': ['lui']}}
        resp = self.app.patch_json(self.group_uri, body,
                                   headers=self.headers)
        storage_size = record_size(self.bucket)
        storage_size += record_size(resp.json['data'])
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 0,
            "record_count": 0,
            "storage_size": storage_size
        })

    def test_tracks_group_delete(self):
        self.create_bucket()
        self.create_group()
        self.app.delete(self.group_uri, headers=self.headers)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 0,
            "record_count": 0,
            "storage_size": record_size(self.bucket)
        })

    #
    # Record
    #

    def test_quota_tracks_record_creation(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        storage_size += record_size(self.record)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 1,
            "storage_size": storage_size
        })

    def test_tracks_record_attributes_update(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        resp = self.app.patch_json(self.record_uri, {'data': {'foo': 'baz'}},
                                   headers=self.headers)
        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        storage_size += record_size(resp.json['data'])
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 1,
            "storage_size": storage_size
        })

    def test_tracks_record_delete(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        self.app.delete(self.record_uri, headers=self.headers)
        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 0,
            "storage_size": storage_size
        })

    def test_tracks_records_delete_with_multiple_records(self):
        self.create_bucket()
        self.create_collection()
        body = {'data': {'foo': 42}}
        self.app.post_json('{}/records'.format(self.collection_uri),
                           body, headers=self.headers)
        self.app.post_json('{}/records'.format(self.collection_uri),
                           body, headers=self.headers)
        self.app.post_json('{}/records'.format(self.collection_uri),
                           body, headers=self.headers)
        self.app.post_json('{}/records'.format(self.collection_uri),
                           body, headers=self.headers)
        self.app.delete('{}/records'.format(self.collection_uri),
                        headers=self.headers)
        storage_size = record_size(self.bucket) + record_size(self.collection)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 0,
            "storage_size": storage_size
        })

    def test_bulk_create(self):
        body = {
            'defaults': {
                'method': 'POST',
                'path': '{}/records'.format(self.collection_uri),
            },
            'requests': [{
                'path': self.bucket_uri,
                'method': 'PUT'
            }, {
                'path': self.collection_uri,
                'method': 'PUT'
            }, {
                'body': {'data': {'id': 'a', 'attr': 1}},
            }, {
                'body': {'data': {'id': 'b', 'attr': 2}},
            }, {
                'body': {'data': {'id': 'c', 'attr': 3}}
            }]
        }
        self.app.post_json('/batch', body, headers=self.headers)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 3,
            "storage_size": 232
        })

    def test_bulk_update(self):
        body = {
            'defaults': {
                'method': 'POST',
                'path': '{}/collections'.format(self.bucket_uri),
            },
            'requests': [{
                'path': self.bucket_uri,
                'method': 'PUT'
            }, {
                'body': {'data': {'id': 'a', 'attr': 10}},
            }, {
                'body': {'data': {'id': 'b', 'attr': 200}},
            }, {
                'body': {'data': {'id': 'c', 'attr': 3000}}
            }]
        }
        self.app.post_json('/batch', body, headers=self.headers)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 3,
            "record_count": 0,
            "storage_size": 196
        })
        body = {
            'defaults': {
                'method': 'PUT',
            },
            'requests': [{
                'path': '{}/collections/a'.format(self.bucket_uri),
                'body': {'data': {'attr': 100}},
            }, {
                'path': '{}/collections/b'.format(self.bucket_uri),
                'body': {'data': {'attr': 2000}},
            }, {
                'path': '{}/collections/c'.format(self.bucket_uri),
                'body': {'data': {'attr': 30000}}
            }]
        }
        self.app.post_json('/batch', body, headers=self.headers)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 3,
            "record_count": 0,
            "storage_size": 199
        })

    def test_bulk_delete(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()

        body = {
            'defaults': {
                'method': 'POST',
                'path': '{}/collections'.format(self.bucket_uri),
            },
            'requests': [{
                'body': {'data': {'id': 'a', 'attr': 1}},
            }, {
                'body': {'data': {'id': 'b', 'attr': 2}},
            }, {
                'body': {'data': {'id': 'c', 'attr': 3}}
            }]
        }
        self.app.post_json('/batch', body, headers=self.headers)

        body = {
            'defaults': {
                'method': 'DELETE',
            },
            'requests': [{
                'path': '{}/collections/a'.format(self.bucket_uri)
            }, {
                'path': '{}/collections/b'.format(self.bucket_uri)
            }, {
                'path': '{}/collections/c'.format(self.bucket_uri)
            }, {
                'path': self.collection_uri
            }]
        }
        self.app.post_json('/batch', body, headers=self.headers)

        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 0,
            "record_count": 0,
            "storage_size": record_size(self.bucket)
        })

        with pytest.raises(RecordNotFoundError):
            self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                             parent_id='{}/collections/a'.format(self.bucket_uri),
                             object_id=QUOTA_COLLECTION_ID)

        with pytest.raises(RecordNotFoundError):
            self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                             parent_id='{}/collections/b'.format(self.bucket_uri),
                             object_id=QUOTA_COLLECTION_ID)

        with pytest.raises(RecordNotFoundError):
            self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                             parent_id='{}/collections/c'.format(self.bucket_uri),
                             object_id=QUOTA_COLLECTION_ID)

        with pytest.raises(RecordNotFoundError):
            self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                             parent_id=self.collection_uri,
                             object_id=QUOTA_COLLECTION_ID)


class QuotaBucketRecordMixin:
    def test_507_is_raised_if_quota_exceeded_on_record_creation(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        body = {'data': {'foo': 42}}
        resp = self.app.post_json('{}/records'.format(self.collection_uri),
                                  body, headers=self.headers, status=507)

        # Check that the storage was not updated.
        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        storage_size += record_size(self.record)

        self.assertFormattedError(
            resp, 507, ERRORS.FORBIDDEN, "Insufficient Storage",
            self.error_message)

        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 1,
            "storage_size": storage_size
        })

        # Check that the record wasn't created
        resp = self.app.get('{}/records'.format(self.collection_uri),
                            headers=self.headers)
        assert len(resp.json['data']) == 1


class QuotaBucketUpdateMixin:
    def test_507_is_raised_if_quota_exceeded_on_record_update(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        body = {'data': {'foo': 42, 'bar': 'This is a very long string.'}}
        resp = self.app.patch_json(self.record_uri,
                                   body, headers=self.headers, status=507)

        # Check that the storage was not updated.
        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        storage_size += record_size(self.record)

        self.assertFormattedError(
            resp, 507, ERRORS.FORBIDDEN, "Insufficient Storage",
            self.error_message)

        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 1,
            "storage_size": storage_size
        })

    def test_507_is_raised_if_quota_exceeded_on_collection_update(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        body = {'data': {'foo': 42, 'bar': 'This is a very long string.'}}
        resp = self.app.patch_json(self.collection_uri,
                                   body, headers=self.headers, status=507)

        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        storage_size += record_size(self.record)

        self.assertFormattedError(
            resp, 507, ERRORS.FORBIDDEN, "Insufficient Storage",
            self.error_message)

        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 1,
            "storage_size": storage_size
        })

    def test_507_is_raised_if_quota_exceeded_on_group_update(self):
        self.create_bucket()
        self.create_collection()
        body = {'data': {'members': []}}
        resp = self.app.put_json(self.group_uri, body,
                                 headers=self.headers)
        group = resp.json['data']
        body = {'data': {'members': ['elle', 'lui', 'je', 'tu', 'il', 'nous',
                                     'vous', 'ils', 'elles']}}
        resp = self.app.put_json(self.group_uri, body,
                                 headers=self.headers, status=507)

        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        storage_size += record_size(group)

        self.assertFormattedError(
            resp, 507, ERRORS.FORBIDDEN, "Insufficient Storage",
            self.error_message)

        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 0,
            "storage_size": storage_size
        })

    def test_507_is_not_raised_if_quota_exceeded_on_record_delete(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        self.app.delete(self.record_uri, headers=self.headers)

        # Check that the storage was not updated.
        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 0,
            "storage_size": storage_size
        })

    def test_507_is_not_raised_if_quota_exceeded_on_collection_delete(self):
        self.create_bucket()
        self.create_collection()
        # fake the quota to the Max
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        data['storage_size'] = 140
        self.storage.update(collection_id=QUOTA_RESOURCE_NAME,
                            parent_id=self.bucket_uri,
                            object_id=QUOTA_BUCKET_ID,
                            record=data)
        transaction.commit()
        self.app.delete(self.collection_uri,
                        headers=self.headers)

        storage_size = 140
        storage_size -= record_size(self.collection)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 0,
            "record_count": 0,
            "storage_size": storage_size
        })

    def test_507_is_raised_if_quota_exceeded_on_group_delete(self):
        self.create_bucket()
        body = {"data": {"members": []}}
        resp = self.app.put_json(self.group_uri, body, headers=self.headers)
        group = resp.json['data']
        # fake the quota to the Max
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        data['storage_size'] = 140
        self.storage.update(collection_id=QUOTA_RESOURCE_NAME,
                            parent_id=self.bucket_uri,
                            object_id=QUOTA_BUCKET_ID,
                            record=data)
        transaction.commit()

        self.app.delete(self.group_uri, headers=self.headers)

        storage_size = 140
        storage_size -= record_size(group)
        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 0,
            "record_count": 0,
            "storage_size": storage_size
        })


class QuotaBucketMixin:
    def test_507_is_raised_if_quota_exceeded_on_collection_creation(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        body = {'data': {'foo': 42}}
        resp = self.app.post_json('{}/collections'.format(self.bucket_uri),
                                  body, headers=self.headers, status=507)

        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        storage_size += record_size(self.record)

        self.assertFormattedError(
            resp, 507, ERRORS.FORBIDDEN, "Insufficient Storage",
            self.error_message)

        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 1,
            "storage_size": storage_size
        })

        # Check that the collection wasn't created
        resp = self.app.get('{}/collections'.format(self.bucket_uri),
                            headers=self.headers)
        assert len(resp.json['data']) == 1

    def test_507_is_raised_if_quota_exceeded_on_group_creation(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        body = {'data': {'members': ['elle']}}
        resp = self.app.put_json(self.group_uri, body,
                                 headers=self.headers, status=507)

        storage_size = record_size(self.bucket)
        storage_size += record_size(self.collection)
        storage_size += record_size(self.record)

        self.assertFormattedError(
            resp, 507, ERRORS.FORBIDDEN, "Insufficient Storage",
            self.error_message)

        data = self.storage.get(collection_id=QUOTA_RESOURCE_NAME,
                                parent_id=self.bucket_uri,
                                object_id=QUOTA_BUCKET_ID)
        self.assertStatsEqual(data, {
            "collection_count": 1,
            "record_count": 1,
            "storage_size": storage_size
        })

        # Check that the group wasn't created
        resp = self.app.get('{}/groups'.format(self.bucket_uri),
                            headers=self.headers)
        assert len(resp.json['data']) == 0


class QuotaMaxBytesExceededSettingsListenerTest(
        FormattedErrorMixin, QuotaBucketRecordMixin, QuotaBucketUpdateMixin,
        QuotaBucketMixin, QuotaWebTest):

    error_message = "Bucket maximum total size exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.bucket_max_bytes'] = '150'
        return settings


class QuotaMaxBytesExceededBucketSettingsListenerTest(
        FormattedErrorMixin, QuotaBucketRecordMixin, QuotaBucketUpdateMixin,
        QuotaBucketMixin, QuotaWebTest):

    error_message = "Bucket maximum total size exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.bucket_test_max_bytes'] = '150'
        return settings


class QuotaMaxItemsExceededSettingsListenerTest(
        FormattedErrorMixin, QuotaBucketRecordMixin, QuotaWebTest):

    error_message = "Bucket maximum number of objects exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.bucket_max_items'] = '1'
        return settings


class QuotaMaxItemsExceededBucketSettingsListenerTest(
        FormattedErrorMixin, QuotaBucketRecordMixin, QuotaWebTest):

    error_message = "Bucket maximum number of objects exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.bucket_test_max_items'] = '1'
        return settings


class QuotaMaxBytesPerItemExceededListenerTest(
        FormattedErrorMixin, QuotaBucketRecordMixin, QuotaBucketUpdateMixin,
        QuotaBucketMixin, QuotaWebTest):

    error_message = "Maximum bytes per object exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.bucket_max_bytes_per_item'] = '55'
        return settings


class QuotaMaxBytesPerItemExceededBucketListenerTest(
        FormattedErrorMixin, QuotaBucketRecordMixin, QuotaBucketUpdateMixin,
        QuotaBucketMixin, QuotaWebTest):

    error_message = "Maximum bytes per object exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.bucket_test_max_bytes_per_item'] = '55'
        return settings


class QuotaCollectionMixin:
    def test_507_is_raised_if_quota_exceeded_on_record_creation(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        body = {'data': {'foo': 42}}
        resp = self.app.post_json('{}/records'.format(self.collection_uri),
                                  body, headers=self.headers, status=507)

        # Check that the storage was not updated.
        storage_size = record_size(self.collection)
        storage_size += record_size(self.record)

        self.assertFormattedError(
            resp, 507, ERRORS.FORBIDDEN, "Insufficient Storage",
            self.error_message)

        data = self.storage.get(QUOTA_RESOURCE_NAME, self.collection_uri,
                                QUOTA_COLLECTION_ID)
        self.assertStatsEqual(data, {
            "record_count": 1,
            "storage_size": storage_size
        })


class QuotaCollectionUpdateMixin:
    def test_507_is_raised_if_quota_exceeded_on_record_update(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        body = {'data': {'foo': 42, 'bar': 'This is a very long string.'}}
        resp = self.app.patch_json(self.record_uri,
                                   body, headers=self.headers, status=507)

        # Check that the storage was not updated.
        storage_size = record_size(self.collection)
        storage_size += record_size(self.record)

        self.assertFormattedError(
            resp, 507, ERRORS.FORBIDDEN, "Insufficient Storage",
            self.error_message)

        data = self.storage.get(QUOTA_RESOURCE_NAME, self.collection_uri,
                                QUOTA_COLLECTION_ID)
        self.assertStatsEqual(data, {
            "record_count": 1,
            "storage_size": storage_size
        })

    def test_507_is_not_raised_if_quota_exceeded_on_record_delete(self):
        self.create_bucket()
        self.create_collection()
        self.create_record()
        self.app.delete(self.record_uri, headers=self.headers)

        # Check that the storage was not updated.
        storage_size = record_size(self.collection)
        data = self.storage.get(QUOTA_RESOURCE_NAME, self.collection_uri,
                                QUOTA_COLLECTION_ID)
        self.assertStatsEqual(data, {
            "record_count": 0,
            "storage_size": storage_size
        })


class QuotaMaxBytesExceededCollectionSettingsListenerTest(
        FormattedErrorMixin, QuotaCollectionMixin, QuotaCollectionUpdateMixin,
        QuotaWebTest):

    error_message = "Collection maximum size exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.collection_max_bytes'] = '100'
        return settings


class QuotaMaxBytesExceededCollectionBucketSettingsListenerTest(
        FormattedErrorMixin, QuotaCollectionMixin, QuotaCollectionUpdateMixin,
        QuotaWebTest):

    error_message = "Collection maximum size exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.collection_test_max_bytes'] = '100'
        return settings


class QuotaMaxBytesExceededBucketCollectionSettingsListenerTest(
        FormattedErrorMixin, QuotaCollectionMixin, QuotaCollectionUpdateMixin,
        QuotaWebTest):

    error_message = "Collection maximum size exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.collection_test_col_max_bytes'] = '100'
        return settings


class QuotaMaxItemsExceededCollectionSettingsListenerTest(
        FormattedErrorMixin, QuotaCollectionMixin, QuotaWebTest):

    error_message = "Collection maximum number of objects exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.collection_max_items'] = '1'
        return settings


class QuotaMaxItemsExceededCollectionBucketSettingsListenerTest(
        FormattedErrorMixin, QuotaCollectionMixin, QuotaWebTest):

    error_message = "Collection maximum number of objects exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.collection_test_max_items'] = '1'
        return settings


class QuotaMaxItemsExceededBucketCollectionSettingsListenerTest(
        FormattedErrorMixin, QuotaCollectionMixin, QuotaWebTest):

    error_message = "Collection maximum number of objects exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.collection_test_col_max_items'] = '1'
        return settings


class QuotaMaxBytesPerItemExceededCollectionSettingsListenerTest(
        FormattedErrorMixin, QuotaCollectionMixin, QuotaWebTest):

    error_message = "Maximum bytes per object exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.collection_max_bytes_per_item'] = '80'
        return settings


class QuotaMaxBytesPerItemExceededCollectionBucketSettingsListenerTest(
        FormattedErrorMixin, QuotaCollectionMixin, QuotaCollectionUpdateMixin,
        QuotaWebTest):

    error_message = "Maximum bytes per object exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.collection_test_max_bytes_per_item'] = '80'
        return settings


class QuotaMaxBytesPerItemExceededBucketCollectionSettingsListenerTest(
        FormattedErrorMixin, QuotaCollectionMixin, QuotaCollectionUpdateMixin,
        QuotaWebTest):

    error_message = "Maximum bytes per object exceeded "

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['quotas.collection_test_col_max_bytes_per_item'] = '80'
        return settings



import mock
import uuid
from contextlib import contextmanager

from pyramid.config import Configurator

from kinto.core import statsd
from kinto.core.events import (ResourceChanged, AfterResourceChanged,
                               ResourceRead, AfterResourceRead, ACTIONS)
from kinto.core.storage.exceptions import BackendError
from kinto.core.testing import unittest

from ..support import BaseWebTest


@contextmanager
def notif_broken(app, event_cls):
    old = app.registry.notify

    def buggy(event):
        if not isinstance(event, event_cls):
            return old(event)
        raise Exception("boom")

    app.registry.notify = buggy
    yield
    app.registry.notify = old


class BaseEventTest(BaseWebTest):

    subscribed = tuple()

    def setUp(self):
        super().setUp()
        self.events = []
        self.body = {'data': {'name': 'de Paris'}}

    def tearDown(self):
        self.events = []
        super().tearDown()

    def listener(self, event):
        self.events.append(event)

    def make_app(self, settings=None, config=None):
        settings = self.get_app_settings(settings)
        config = Configurator(settings=settings)
        for event_cls in self.subscribed:
            config.add_subscriber(self.listener, event_cls)
        config.commit()
        return super().make_app(settings=settings, config=config)


class ResourceReadTest(BaseEventTest, unittest.TestCase):

    subscribed = (ResourceRead,)

    def test_get_sends_read_event(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record_id = resp.json['data']['id']
        record_url = self.get_item_url(record_id)
        self.app.get(record_url, headers=self.headers)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'], ACTIONS.READ.value)
        self.assertEqual(len(self.events[0].read_records), 1)

    def test_collection_get_sends_read_event(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'], ACTIONS.READ.value)
        self.assertEqual(len(self.events[0].read_records), 0)

    def test_post_sends_read_if_id_already_exists(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        body = dict(self.body)
        body['data']['id'] = record['id']

        # a second post with the same record id
        self.app.post_json(self.collection_url, body, headers=self.headers,
                           status=200)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'], ACTIONS.READ.value)


class ResourceChangedTest(BaseEventTest, unittest.TestCase):

    subscribed = (ResourceChanged,)

    def test_events_have_custom_representation(self):
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)
        self.assertEqual(repr(self.events[0]),
                         "<ResourceChanged action=create "
                         "uri={}>".format(self.collection_url))

    def test_post_sends_create_action(self):
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)

    def test_put_sends_create_action(self):
        body = dict(self.body)
        body['data']['id'] = record_id = str(uuid.uuid4())
        record_url = self.get_item_url(record_id)
        self.app.put_json(record_url, body,
                          headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)

    def test_not_triggered_on_failed_put(self):
        record_id = str(uuid.uuid4())
        record_url = self.get_item_url(record_id)
        self.app.put_json(record_url, self.body, headers=self.headers)
        headers = {**self.headers, 'If-Match': '"12345"'}
        self.app.put_json(record_url, self.body, headers=headers, status=412)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)

    def test_patch_sends_update_action(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        record_url = self.get_item_url(record['id'])

        self.app.patch_json(record_url, self.body, headers=self.headers,
                            status=200)
        self.assertEqual(len(self.events), 2)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload['action'],
                         ACTIONS.UPDATE.value)

    def test_put_sends_update_action_if_record_exists(self):
        body = dict(self.body)
        body['data']['id'] = record_id = str(uuid.uuid4())
        record_url = self.get_item_url(record_id)
        self.app.put_json(record_url, body,
                          headers=self.headers, status=201)

        body['data']['more'] = 'stuff'
        self.app.put_json(record_url, body,
                          headers=self.headers, status=200)

        self.assertEqual(len(self.events), 2)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload['action'],
                         ACTIONS.UPDATE.value)

    def test_delete_sends_delete_action(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        record_url = self.get_item_url(record['id'])

        self.app.delete(record_url, headers=self.headers, status=200)
        self.assertEqual(len(self.events), 2)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload['action'],
                         ACTIONS.DELETE.value)

    def test_collection_delete_sends_delete_action(self):
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)
        self.app.post_json(self.collection_url, self.body,
                           headers=self.headers, status=201)

        self.app.delete(self.collection_url, headers=self.headers, status=200)

        self.assertEqual(len(self.events), 3)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[1].payload['action'],
                         ACTIONS.CREATE.value)
        self.assertEqual(self.events[2].payload['action'],
                         ACTIONS.DELETE.value)

    def test_request_fails_if_notify_fails(self):
        with notif_broken(self.app.app, ResourceChanged):
            self.app.post_json(self.collection_url, self.body,
                               headers=self.headers, status=500)
        self.assertEqual(len(self.events), 0)

    def test_triggered_on_protected_resource(self):
        app = self.make_app(settings={
            'psilo_write_principals': 'system.Authenticated'
        })
        app.post_json('/psilos', self.body,
                      headers=self.headers, status=201)
        self.assertEqual(len(self.events), 1)
        self.assertEqual(self.events[0].payload['action'],
                         ACTIONS.CREATE.value)

    def test_permissions_are_stripped_from_event_on_protected_resource(self):
        app = self.make_app(settings={
            'psilo_write_principals': 'system.Authenticated'
        })
        resp = app.post_json('/psilos', self.body,
                             headers=self.headers, status=201)
        record = resp.json['data']
        record_url = '/psilos/{}'.format(record['id'])
        app.patch_json(record_url, {"data": {"name": "De barcelona"}},
                       headers=self.headers)
        impacted_records = self.events[-1].impacted_records
        self.assertNotIn('__permissions__', impacted_records[0]['new'])
        self.assertNotIn('__permissions__', impacted_records[0]['old'])


class AfterResourceChangedTest(BaseEventTest, unittest.TestCase):

    subscribed = (AfterResourceChanged,)

    def test_request_succeeds_if_notify_fails(self):
        with notif_broken(self.app.app, AfterResourceChanged):
            self.app.post_json(self.collection_url, self.body,
                               headers=self.headers)

        self.assertEqual(len(self.events), 0)


class AfterResourceReadTest(BaseEventTest, unittest.TestCase):

    subscribed = (AfterResourceRead,)

    def test_request_succeeds_if_notify_fails(self):
        with notif_broken(self.app.app, AfterResourceChanged):
            self.app.post_json(self.collection_url, self.body,
                               headers=self.headers)

        self.assertEqual(len(self.events), 0)


class ImpactedRecordsTest(BaseEventTest, unittest.TestCase):

    subscribed = (ResourceChanged,)

    def test_create_has_new_record_and_no_old_in_payload(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers)
        record = resp.json['data']
        impacted_records = self.events[-1].impacted_records
        self.assertEqual(len(impacted_records), 1)
        self.assertNotIn('old', impacted_records[0])
        self.assertEqual(impacted_records[0]['new'], record)

    def test_collection_delete_has_old_and_new_in_payload(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers)
        record1 = resp.json['data']
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers)
        record2 = resp.json['data']

        self.app.delete(self.collection_url, headers=self.headers, status=200)

        impacted_records = self.events[-1].impacted_records
        self.assertEqual(len(impacted_records), 2)
        impacted_records = sorted(impacted_records,
                                  key=lambda x: x['old']['last_modified'])
        self.assertEqual(impacted_records[0]['old'], record1)
        self.assertEqual(impacted_records[1]['old'], record2)
        self.assertEqual(impacted_records[0]['new']['deleted'], True)
        self.assertEqual(impacted_records[1]['new']['deleted'], True)
        self.assertEqual(impacted_records[0]['old']['id'],
                         impacted_records[0]['new']['id'])
        self.assertEqual(impacted_records[1]['old']['id'],
                         impacted_records[1]['new']['id'])
        deleted_ids = {impacted_records[0]['new']['id'],
                       impacted_records[1]['new']['id']}
        self.assertEqual(deleted_ids, {record1['id'], record2['id']})

    def test_update_has_old_and_new_record(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        record_url = self.get_item_url(record['id'])
        self.app.patch_json(record_url, {'data': {'name': 'en boite'}},
                            headers=self.headers)
        impacted_records = self.events[-1].impacted_records
        self.assertEqual(len(impacted_records), 1)
        self.assertEqual(impacted_records[0]['new']['id'], record['id'])
        self.assertEqual(impacted_records[0]['new']['id'],
                         impacted_records[0]['old']['id'])
        self.assertEqual(impacted_records[0]['old']['name'], 'de Paris')
        self.assertEqual(impacted_records[0]['new']['name'], 'en boite')

    def test_delete_has_old_and_new_record_in_payload(self):
        resp = self.app.post_json(self.collection_url, self.body,
                                  headers=self.headers, status=201)
        record = resp.json['data']
        record_url = self.get_item_url(record['id'])
        self.app.delete(record_url, headers=self.headers, status=200)

        impacted_records = self.events[-1].impacted_records
        self.assertEqual(len(impacted_records), 1)
        self.assertEqual(impacted_records[0]['old'], record)
        self.assertEqual(impacted_records[0]['new']['id'], record['id'])
        self.assertEqual(impacted_records[0]['new']['deleted'], True)


class BatchEventsTest(BaseEventTest, unittest.TestCase):

    subscribed = (ResourceChanged, ResourceRead)

    def test_impacted_records_are_merged(self):
        record_id = str(uuid.uuid4())
        record_url = self.get_item_url(record_id)
        body = {
            "defaults": {
                "method": "PUT",
                "path": record_url
            },
            "requests": [
                {"body": {'data': {'name': 'foo'}}},
                {"body": {'data': {'name': 'bar'}}},
                {"body": {'data': {'name': 'baz'}}},
                {"method": "DELETE"}
            ]
        }
        self.app.post_json("/batch", body, headers=self.headers)
        self.assertEqual(len(self.events), 3)

        create_event = self.events[0]
        self.assertEqual(create_event.payload['action'], 'create')
        self.assertEqual(len(create_event.impacted_records), 1)
        self.assertNotIn('old', create_event.impacted_records[0])
        update_event = self.events[1]
        self.assertEqual(update_event.payload['action'], 'update')
        impacted = update_event.impacted_records
        self.assertEqual(len(impacted), 2)
        self.assertEqual(impacted[0]['old']['name'], 'foo')
        self.assertEqual(impacted[0]['new']['name'], 'bar')
        self.assertEqual(impacted[1]['old']['name'], 'bar')
        self.assertEqual(impacted[1]['new']['name'], 'baz')
        delete_event = self.events[2]
        self.assertEqual(delete_event.payload['action'], 'delete')
        self.assertEqual(len(delete_event.impacted_records), 1)
        self.assertIn('old', delete_event.impacted_records[0])
        self.assertIn('new', delete_event.impacted_records[0])

    def test_one_event_is_sent_per_resource(self):
        body = {
            "defaults": {
                "method": "POST",
                "body": self.body,
            },
            "requests": [
                {"path": '/mushrooms'},
                {"path": '/mushrooms'},
                {"path": '/psilos'},
            ]
        }
        self.app.post_json("/batch", body, headers=self.headers)
        self.assertEqual(len(self.events), 2)

    def test_one_event_is_sent_per_parent_id(self):
        # /mushrooms is a UserResource (see testapp.views), which means
        # that parent_id depends on the authenticated user.
        body = {
            "defaults": {
                "path": '/mushrooms',
                "method": "POST",
                "body": self.body
            },
            "requests": [
                {"headers": {"Authorization": "Basic bWF0OjE="}},
                {"headers": {"Authorization": "Basic dG90bzp0dXR1"}},
                {"headers": {"Authorization": "Basic bWF0OjE="}},
            ]
        }
        self.app.post_json("/batch", body, headers=self.headers)
        # Two different auth headers, thus two different parent_id:
        self.assertEqual(len(self.events), 2)

    def test_one_event_is_sent_per_action(self):
        body = {
            "defaults": {
                "path": '/mushrooms',
            },
            "requests": [
                {"method": "POST", "body": self.body},
                {"method": "DELETE"},
                {"method": "GET"},
            ]
        }
        self.app.post_json("/batch", body, headers=self.headers)
        self.assertEqual(len(self.events), 3)

    def test_events_are_not_sent_if_subrequest_fails(self):
        patch = mock.patch.object(self.storage,
                                  'delete_all',
                                  side_effect=BackendError('boom'))
        patch.start()
        self.addCleanup(patch.stop)
        request_create = {
            "method": "POST",
            "body": self.body,
        }
        request_delete_all = {
            "method": "DELETE",
            "body": self.body,
        }
        body = {
            "defaults": {
                "path": self.collection_url
            },
            "requests": [request_create, request_delete_all]
        }
        self.app.post_json("/batch", body, headers=self.headers,
                           status=503)
        self.assertEqual(len(self.events), 0)


def load_from_config(config, prefix):
    class ClassListener:
        def __call__(self, event):
            pass
    return ClassListener()


@unittest.skipIf(not statsd.statsd_module, "statsd is not installed.")
class StatsDTest(BaseWebTest, unittest.TestCase):
    def get_app_settings(self, *args, **kwargs):
        settings = super().get_app_settings(*args, **kwargs)
        if not statsd.statsd_module:
            return settings

        settings['statsd_url'] = 'udp://localhost:8125'
        this_module = 'tests.core.resource.test_events'
        settings['event_listeners'] = 'test'
        settings['event_listeners.test.use'] = this_module
        return settings

    def test_statds_tracks_listeners_execution_duration(self):
        statsd_client = self.app.app.registry.statsd._client
        with mock.patch.object(statsd_client, 'timing') as mocked:
            self.app.post_json(self.collection_url,
                               {"data": {"name": "pouet"}},
                               headers=self.headers)
            timers = set(c[0][0] for c in mocked.call_args_list)
            self.assertIn('listeners.test', timers)



import builtins
import logging
import mock
import os
import pytest
import sys
import tempfile
import unittest

from io import StringIO

from kinto import __version__ as kinto_version
from kinto.__main__ import main, DEFAULT_LOG_FORMAT

_, TEMP_KINTO_INI = tempfile.mkstemp(prefix='kinto_config', suffix='.ini')


class TestMain(unittest.TestCase):
    def setUp(self):
        try:
            os.remove(TEMP_KINTO_INI)
        except OSError:
            pass

    def test_cli_init_generates_configuration(self):
        res = main(['init', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
        assert res == 0
        assert os.path.exists(TEMP_KINTO_INI)

    def test_cli_init_returns_if_file_exists(self):
        with open(TEMP_KINTO_INI, 'w') as f:
            f.write("exists")
        with mock.patch('sys.stderr', new_callable=StringIO) as mock_stderr:
            res = main(['init', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            assert res == 1
            content = mock_stderr.getvalue()
            assert '{} already exists.'.format(TEMP_KINTO_INI) in content

    def test_cli_init_asks_for_backend_if_not_specified(self):
        with mock.patch("kinto.__main__.input", create=True, return_value="2"):
            res = main(['init', '--ini', TEMP_KINTO_INI])
            assert res == 0
        with open(TEMP_KINTO_INI) as f:
            content = f.read()
        assert 'redis' in content

    def test_cli_init_asks_until_backend_is_valid(self):
        with mock.patch("kinto.__main__.input", create=True, side_effect=["10", "2"]):
            res = main(['init', '--ini', TEMP_KINTO_INI])
            assert res == 0
            with open(TEMP_KINTO_INI) as f:
                content = f.read()
            assert 'redis' in content

    def test_fails_if_not_enough_args(self):
        with mock.patch('sys.stderr', new_callable=StringIO) as mock_stderr:
            with pytest.raises(SystemExit) as excinfo:
                main([])
            assert 'arguments are required: subcommand' in mock_stderr.getvalue()
            assert excinfo.value.code == 2

    def test_cli_init_installs_postgresql_dependencies_if_needed(self):
        realimport = builtins.__import__

        def psycopg2_missing(name, *args, **kwargs):
            if name == 'psycopg2':
                raise ImportError()
            else:
                return realimport(name, *args, **kwargs)

        with mock.patch('builtins.__import__', side_effect=psycopg2_missing):
            with mock.patch('pip.main', return_value=None) as mocked_pip:
                with mock.patch("kinto.__main__.input", create=True, return_value="1"):
                    res = main(['init', '--ini', TEMP_KINTO_INI])
                    assert res == 0
                    assert mocked_pip.call_count == 1

    def test_cli_init_installs_redis_dependencies_if_needed(self):
        realimport = builtins.__import__

        def redis_missing(name, *args, **kwargs):
            if name == 'kinto_redis':
                raise ImportError()
            else:
                return realimport(name, *args, **kwargs)

        with mock.patch('builtins.__import__', side_effect=redis_missing):
            with mock.patch('pip.main', return_value=None) as mocked_pip:
                with mock.patch("kinto.__main__.input", create=True, return_value="2"):
                    res = main(['init', '--ini', TEMP_KINTO_INI])
                    assert res == 0
                    assert mocked_pip.call_count == 1

    def test_main_takes_sys_argv_by_default(self):
        testargs = ['prog', 'init', '--ini', TEMP_KINTO_INI, '--backend', 'memory']
        with mock.patch.object(sys, 'argv', testargs):
            main()

        with open(TEMP_KINTO_INI) as f:
            content = f.read()
        assert 'memory' in content

    def test_cli_migrate_command_runs_init_schema(self):
        with mock.patch('kinto.__main__.scripts.migrate') as mocked_migrate:
            res = main(['init', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            assert res == 0
            res = main(['migrate', '--ini', TEMP_KINTO_INI])
            assert res == 0
            assert mocked_migrate.call_count == 1

    def test_cli_delete_collection_run_delete_collection_script(self):
        with mock.patch('kinto.__main__.scripts.delete_collection') as del_col:
            del_col.return_value = mock.sentinel.del_col_code
            res = main(['init', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            assert res == 0
            res = main(['delete-collection', '--ini', TEMP_KINTO_INI,
                        '--bucket', 'test_bucket',
                        '--collection', 'test_collection'])
            assert res == mock.sentinel.del_col_code
            assert del_col.call_count == 1

    def test_cli_start_runs_pserve(self):
        with mock.patch('kinto.__main__.pserve.main') as mocked_pserve:
            res = main(['init', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            assert res == 0
            res = main(['start', '--ini', TEMP_KINTO_INI])
            assert res == 0
            assert mocked_pserve.call_count == 1

    def test_cli_start_with_quiet_option_runs_pserve_with_quiet(self):
        with mock.patch('kinto.__main__.pserve.main') as mocked_pserve:
            res = main(['init', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            assert res == 0
            res = main(['start', '-q', '--ini', TEMP_KINTO_INI])
            assert res == 0
            assert mocked_pserve.call_count == 1
            assert mocked_pserve.call_args[1]['argv'][1] == '-q'

    def test_cli_start_with_verbose_option_runs_pserve_with_verbose(self):
        with mock.patch('kinto.__main__.pserve.main') as mocked_pserve:
            res = main(['init', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            assert res == 0
            res = main(['start', '-v', '--ini', TEMP_KINTO_INI])
            assert res == 0
            assert mocked_pserve.call_count == 1
            assert mocked_pserve.call_args[1]['argv'][1] == '-v'

    def test_cli_start_with_reload_runs_pserve_with_reload(self):
        with mock.patch('kinto.__main__.pserve.main') as mocked_pserve:
            res = main(['init', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            assert res == 0
            res = main(['start', '--reload', '--ini', TEMP_KINTO_INI])
            assert res == 0
            assert mocked_pserve.call_count == 1
            assert '--reload' in mocked_pserve.call_args[1]['argv']

    def test_cli_can_display_kinto_version(self):
        with mock.patch('sys.stdout', new_callable=StringIO) as mock_stdout:
            res = main(['version'])
            assert mock_stdout.getvalue() == '{}\n'.format(kinto_version)
            assert res == 0

    def test_cli_can_configure_logger_in_quiet(self):
        with mock.patch('kinto.__main__.logging') as mocked_logging:
            main(['init', '-q', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            mocked_logging.basicConfig.assert_called_with(
                level=mocked_logging.CRITICAL,
                format=DEFAULT_LOG_FORMAT)

    def test_cli_can_configure_logger_in_debug(self):
        with mock.patch('kinto.__main__.logging') as mocked_logging:
            main(['init', '--debug', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            mocked_logging.basicConfig.assert_called_with(
                level=mocked_logging.DEBUG,
                format=DEFAULT_LOG_FORMAT)

    def test_cli_use_default_logging_logger(self):
        with mock.patch('kinto.__main__.logging') as mocked_logging:
            main(['init', '--ini', TEMP_KINTO_INI, '--backend', 'memory'])
            mocked_logging.basicConfig.assert_called_with(
                level=logging.INFO,
                format=DEFAULT_LOG_FORMAT)



import re
import functools

import colander
import venusian

from pyramid import exceptions as pyramid_exceptions
from pyramid.decorator import reify
from pyramid.security import Everyone
from pyramid.httpexceptions import (HTTPNotModified, HTTPPreconditionFailed,
                                    HTTPNotFound, HTTPServiceUnavailable)

from kinto.core import logger
from kinto.core import Service
from kinto.core.errors import http_error, raise_invalid, send_alert, ERRORS, request_GET
from kinto.core.events import ACTIONS
from kinto.core.storage import exceptions as storage_exceptions, Filter, Sort
from kinto.core.utils import (
    COMPARISON, classname, decode64, encode64, json, find_nested_value,
    dict_subset, recursive_update_dict, apply_json_patch
)

from .model import Model, ShareableModel
from .schema import ResourceSchema, JsonPatchRequestSchema
from .viewset import ViewSet, ShareableViewSet


def register(depth=1, **kwargs):
    """Ressource class decorator.

    Register the decorated class in the cornice registry.
    Pass all its keyword arguments to the register_resource
    function.
    """
    def wrapped(resource):
        register_resource(resource, depth=depth + 1, **kwargs)
        return resource
    return wrapped


def register_resource(resource_cls, settings=None, viewset=None, depth=1,
                      **kwargs):
    """Register a resource in the cornice registry.

    :param resource_cls:
        The resource class to register.
        It should be a class or have a "name" attribute.

    :param viewset:
        A ViewSet object, which will be used to find out which arguments should
        be appended to the views, and where the views are.

    :param depth:
        A depth offset. It will be used to determine what is the level of depth
        in the call tree. (set to 1 by default.)

    Any additional keyword parameters will be used to override the viewset
    attributes.
    """
    if viewset is None:
        viewset = resource_cls.default_viewset(**kwargs)
    else:
        viewset.update(**kwargs)

    resource_name = viewset.get_name(resource_cls)

    def register_service(endpoint_type, settings):
        """Registers a service in cornice, for the given type.
        """
        path_pattern = getattr(viewset, '{}_path'.format(endpoint_type))
        path_values = {'resource_name': resource_name}
        path = path_pattern.format_map(path_values)

        name = viewset.get_service_name(endpoint_type, resource_cls)

        service = Service(name, path, depth=depth,
                          **viewset.get_service_arguments())

        # Attach viewset and resource to the service for later reference.
        service.viewset = viewset
        service.resource = resource_cls
        service.type = endpoint_type
        # Attach collection and record paths.
        service.collection_path = viewset.collection_path.format_map(path_values)
        service.record_path = (viewset.record_path.format_map(path_values)
                               if viewset.record_path is not None else None)

        methods = getattr(viewset, '{}_methods'.format(endpoint_type))
        for method in methods:
            if not viewset.is_endpoint_enabled(
                    endpoint_type, resource_name, method.lower(), settings):
                continue

            argument_getter = getattr(viewset, '{}_arguments'.format(endpoint_type))
            view_args = argument_getter(resource_cls, method)

            view = viewset.get_view(endpoint_type, method.lower())
            service.add_view(method, view, klass=resource_cls, **view_args)

            # We support JSON-patch on PATCH views. Since the body payload
            # of JSON Patch is not a dict (mapping) but an array, we can't
            # use the same schema as for other PATCH protocols. We add another
            # dedicated view for PATCH, but targetting a different content_type
            # predicate.
            if method.lower() == "patch":
                view_args['content_type'] = "application/json-patch+json"
                view_args['schema'] = JsonPatchRequestSchema()
                service.add_view(method, view, klass=resource_cls, **view_args)

        return service

    def callback(context, name, ob):
        # get the callbacks registred by the inner services
        # and call them from here when the @resource classes are being
        # scanned by venusian.
        config = context.config.with_package(info.module)

        # Storage is mandatory for resources.
        if not hasattr(config.registry, 'storage'):
            msg = 'Mandatory storage backend is missing from configuration.'
            raise pyramid_exceptions.ConfigurationError(msg)

        # A service for the list.
        service = register_service('collection', config.registry.settings)
        config.add_cornice_service(service)
        # An optional one for record endpoint.
        if getattr(viewset, 'record_path') is not None:
            service = register_service('record', config.registry.settings)
            config.add_cornice_service(service)

    info = venusian.attach(resource_cls, callback, category='pyramid', depth=depth)
    return callback


class UserResource:
    """Base resource class providing every endpoint."""

    default_viewset = ViewSet
    """Default :class:`kinto.core.resource.viewset.ViewSet` class to use when
    the resource is registered."""

    default_model = Model
    """Default :class:`kinto.core.resource.model.Model` class to use for
    interacting the :mod:`kinto.core.storage` and :mod:`kinto.core.permission`
    backends."""

    schema = ResourceSchema
    """Schema to validate records."""

    def __init__(self, request, context=None):
        # Models are isolated by user.
        parent_id = self.get_parent_id(request)

        # Authentication to storage is transmitted as is (cf. cloud_storage).
        auth = request.headers.get('Authorization')

        # ID generator by resource name in settings.
        default_id_generator = request.registry.id_generators['']
        resource_name = context.resource_name if context else ''
        id_generator = request.registry.id_generators.get(resource_name,
                                                          default_id_generator)

        self.model = self.default_model(
            storage=request.registry.storage,
            id_generator=id_generator,
            collection_id=classname(self),
            parent_id=parent_id,
            auth=auth)

        self.request = request
        self.context = context
        self.record_id = self.request.matchdict.get('id')
        self.force_patch_update = False

        content_type = str(self.request.headers.get('Content-Type')).lower()
        self._is_json_patch = content_type == 'application/json-patch+json'

        # Log resource context.
        logger.bind(collection_id=self.model.collection_id,
                    collection_timestamp=self.timestamp)

    @reify
    def timestamp(self):
        """Return the current collection timestamp.

        :rtype: int
        """
        try:
            return self.model.timestamp()
        except storage_exceptions.BackendError as e:
            is_readonly = self.request.registry.settings['readonly']
            if not is_readonly:
                raise e
            # If the instance is configured to be readonly, and if the
            # collection is empty, the backend will try to bump the timestamp.
            # It fails if the configured db user has not write privileges.
            logger.exception(e)
            error_msg = ("Collection timestamp cannot be written. "
                         "Records endpoint must be hit at least once from a "
                         "writable instance.")
            raise http_error(HTTPServiceUnavailable(),
                             errno=ERRORS.BACKEND,
                             message=error_msg)

    def get_parent_id(self, request):
        """Return the parent_id of the resource with regards to the current
        request.

        :param request:
            The request used to create the resource.

        :rtype: str

        """
        return request.prefixed_userid

    def _get_known_fields(self):
        """Return all the `field` defined in the ressource schema."""
        known_fields = [c.name for c in self.schema().children] + \
                       [self.model.id_field,
                        self.model.modified_field,
                        self.model.deleted_field]
        return known_fields

    def is_known_field(self, field):
        """Return ``True`` if `field` is defined in the resource schema.
        If the resource schema allows unknown fields, this will always return
        ``True``.

        :param str field: Field name
        :rtype: bool

        """
        if self.schema.get_option('preserve_unknown'):
            return True

        known_fields = self._get_known_fields()
        # Test first level only: ``target.data.id`` -> ``target``
        field = field.split('.', 1)[0]
        return field in known_fields

    #
    # End-points
    #

    def collection_get(self):
        """Model ``GET`` endpoint: retrieve multiple records.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified` if
            ``If-None-Match`` header is provided and collection not
            modified in the interim.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and collection modified
            in the iterim.
        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if filters or sorting are invalid.
        """
        self._add_timestamp_header(self.request.response)
        self._add_cache_header(self.request.response)
        self._raise_304_if_not_modified()
        # Collections are considered resources that always exist
        self._raise_412_if_modified(record={})

        headers = self.request.response.headers

        filters = self._extract_filters()
        limit = self._extract_limit()
        sorting = self._extract_sorting(limit)
        partial_fields = self._extract_partial_fields()

        filter_fields = [f.field for f in filters]
        include_deleted = self.model.modified_field in filter_fields

        pagination_rules, offset = self._extract_pagination_rules_from_token(
            limit, sorting)

        records, total_records = self.model.get_records(
            filters=filters,
            sorting=sorting,
            limit=limit,
            pagination_rules=pagination_rules,
            include_deleted=include_deleted)

        offset = offset + len(records)
        if limit and len(records) == limit and offset < total_records:
            lastrecord = records[-1]
            next_page = self._next_page_url(sorting, limit, lastrecord, offset)
            headers['Next-Page'] = next_page

        if partial_fields:
            records = [
                dict_subset(record, partial_fields)
                for record in records
            ]

        # Bind metric about response size.
        logger.bind(nb_records=len(records), limit=limit)
        headers['Total-Records'] = str(total_records)

        return self.postprocess(records)

    def collection_post(self):
        """Model ``POST`` endpoint: create a record.

        If the new record id conflicts against an existing one, the
        posted record is ignored, and the existing record is returned, with
        a ``200`` status.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and collection modified
            in the iterim.

        .. seealso::

            Add custom behaviour by overriding
            :meth:`kinto.core.resource.UserResource.process_record`
        """
        new_record = self.request.validated['body'].get('data', {})
        try:
            # Since ``id`` does not belong to schema, it is not in validated
            # data. Must look up in body.
            id_field = self.model.id_field
            new_record[id_field] = _id = self.request.json['data'][id_field]
            self._raise_400_if_invalid_id(_id)
            existing = self._get_record_or_404(_id)
        except (HTTPNotFound, KeyError, ValueError):
            existing = None

        self._raise_412_if_modified(record=existing)

        if existing:
            record = existing
            action = ACTIONS.READ
        else:
            new_record = self.process_record(new_record)
            record = self.model.create_record(new_record)
            self.request.response.status_code = 201
            action = ACTIONS.CREATE

        timestamp = record[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        return self.postprocess(record, action=action)

    def collection_delete(self):
        """Model ``DELETE`` endpoint: delete multiple records.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and collection modified
            in the iterim.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if filters are invalid.
        """
        # Collections are considered resources that always exist
        self._raise_412_if_modified(record={})

        filters = self._extract_filters()
        limit = self._extract_limit()
        sorting = self._extract_sorting(limit)
        pagination_rules, offset = self._extract_pagination_rules_from_token(limit, sorting)

        records, total_records = self.model.get_records(filters=filters,
                                                        sorting=sorting,
                                                        limit=limit,
                                                        pagination_rules=pagination_rules)
        deleted = self.model.delete_records(filters=filters,
                                            sorting=sorting,
                                            limit=limit,
                                            pagination_rules=pagination_rules)
        if deleted:
            lastrecord = deleted[-1]
            # Get timestamp of the last deleted field
            timestamp = lastrecord[self.model.modified_field]
            self._add_timestamp_header(self.request.response, timestamp=timestamp)

            # Add pagination header
            offset = offset + len(deleted)
            if limit and len(deleted) == limit and offset < total_records:
                next_page = self._next_page_url(sorting, limit, lastrecord, offset)
                self.request.response.headers['Next-Page'] = next_page
        else:
            self._add_timestamp_header(self.request.response)

        headers = self.request.response.headers
        headers['Total-Records'] = str(total_records)

        action = len(deleted) > 0 and ACTIONS.DELETE or ACTIONS.READ
        return self.postprocess(deleted, action=action, old=records)

    def get(self):
        """Record ``GET`` endpoint: retrieve a record.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the record is not found.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified` if
            ``If-None-Match`` header is provided and record not
            modified in the interim.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and record modified
            in the iterim.
        """
        self._raise_400_if_invalid_id(self.record_id)
        record = self._get_record_or_404(self.record_id)
        timestamp = record[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)
        self._add_cache_header(self.request.response)
        self._raise_304_if_not_modified(record)
        self._raise_412_if_modified(record)

        partial_fields = self._extract_partial_fields()
        if partial_fields:
            record = dict_subset(record, partial_fields)

        return self.postprocess(record)

    def put(self):
        """Record ``PUT`` endpoint: create or replace the provided record and
        return it.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and record modified
            in the iterim.

        .. note::

            If ``If-None-Match: *`` request header is provided, the
            ``PUT`` will succeed only if no record exists with this id.

        .. seealso::

            Add custom behaviour by overriding
            :meth:`kinto.core.resource.UserResource.process_record`.
        """
        self._raise_400_if_invalid_id(self.record_id)
        try:
            existing = self._get_record_or_404(self.record_id)
        except HTTPNotFound:
            existing = None

        self._raise_412_if_modified(record=existing)

        # If `data` is not provided, use existing record (or empty if creation)
        post_record = self.request.validated['body'].get('data', existing) or {}

        record_id = post_record.setdefault(self.model.id_field, self.record_id)
        self._raise_400_if_id_mismatch(record_id, self.record_id)

        new_record = self.process_record(post_record, old=existing)

        if existing:
            record = self.model.update_record(new_record)
        else:
            record = self.model.create_record(new_record)
            self.request.response.status_code = 201

        timestamp = record[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        action = existing and ACTIONS.UPDATE or ACTIONS.CREATE
        return self.postprocess(record, action=action, old=existing)

    def patch(self):
        """Record ``PATCH`` endpoint: modify a record and return its
        new version.

        If a request header ``Response-Behavior`` is set to ``light``,
        only the fields whose value was changed are returned.
        If set to ``diff``, only the fields whose value became different than
        the one provided are returned.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the record is not found.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and record modified
            in the iterim.

        .. seealso::
            Add custom behaviour by overriding
            :meth:`kinto.core.resource.UserResource.apply_changes` or
            :meth:`kinto.core.resource.UserResource.process_record`.
        """
        self._raise_400_if_invalid_id(self.record_id)
        existing = self._get_record_or_404(self.record_id)
        self._raise_412_if_modified(existing)

        # patch is specified as a list of of operations (RFC 6902)
        if self._is_json_patch:
            requested_changes = self.request.validated['body']
        else:
            # `data` attribute may not be present if only perms are patched.
            body = self.request.validated['body']
            if not body:
                # If no `data` nor `permissions` is provided in patch, reject!
                # XXX: This should happen in schema instead (c.f. ShareableViewSet)
                error_details = {
                    'name': 'data',
                    'description': 'Provide at least one of data or permissions',
                }
                raise_invalid(self.request, **error_details)
            requested_changes = body.get('data', {})

        updated, applied_changes = self.apply_changes(existing,
                                                      requested_changes=requested_changes)

        record_id = updated.setdefault(self.model.id_field,
                                       self.record_id)
        self._raise_400_if_id_mismatch(record_id, self.record_id)

        new_record = self.process_record(updated, old=existing)

        changed_fields = [k for k in applied_changes.keys()
                          if existing.get(k) != new_record.get(k)]

        # Save in storage if necessary.
        if changed_fields or self.force_patch_update:
            new_record = self.model.update_record(new_record)

        else:
            # Behave as if storage would have added `id` and `last_modified`.
            for extra_field in [self.model.modified_field,
                                self.model.id_field]:
                new_record[extra_field] = existing[extra_field]

        # Adjust response according to ``Response-Behavior`` header
        body_behavior = self.request.validated['header'].get('Response-Behavior', 'full')

        if body_behavior.lower() == 'light':
            # Only fields that were changed.
            data = {k: new_record[k] for k in changed_fields}

        elif body_behavior.lower() == 'diff':
            # Only fields that are different from those provided.
            data = {k: new_record[k] for k in changed_fields
                    if applied_changes.get(k) != new_record.get(k)}
        else:
            data = new_record

        timestamp = new_record.get(self.model.modified_field,
                                   existing[self.model.modified_field])
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        return self.postprocess(data, action=ACTIONS.UPDATE, old=existing)

    def delete(self):
        """Record ``DELETE`` endpoint: delete a record and return it.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the record is not found.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed` if
            ``If-Match`` header is provided and record modified
            in the iterim.
        """
        self._raise_400_if_invalid_id(self.record_id)
        record = self._get_record_or_404(self.record_id)
        self._raise_412_if_modified(record)

        # Retreive the last_modified information from a querystring if present.
        last_modified = self.request.validated['querystring'].get('last_modified')

        # If less or equal than current record. Ignore it.
        if last_modified and last_modified <= record[self.model.modified_field]:
            last_modified = None

        deleted = self.model.delete_record(record, last_modified=last_modified)
        timestamp = deleted[self.model.modified_field]
        self._add_timestamp_header(self.request.response, timestamp=timestamp)

        return self.postprocess(deleted, action=ACTIONS.DELETE, old=record)

    #
    # Data processing
    #

    def process_record(self, new, old=None):
        """Hook for processing records before they reach storage, to introduce
        specific logics on fields for example.

        .. code-block:: python

            def process_record(self, new, old=None):
                new = super().process_record(new, old)
                version = old['version'] if old else 0
                new['version'] = version + 1
                return new

        Or add extra validation based on request:

        .. code-block:: python

            from kinto.core.errors import raise_invalid

            def process_record(self, new, old=None):
                new = super().process_record(new, old)
                if new['browser'] not in request.headers['User-Agent']:
                    raise_invalid(self.request, name='browser', error='Wrong')
                return new

        :param dict new: the validated record to be created or updated.
        :param dict old: the old record to be updated,
            ``None`` for creation endpoints.

        :returns: the processed record.
        :rtype: dict
        """
        modified_field = self.model.modified_field
        new_last_modified = new.get(modified_field)

        # Drop the new last_modified if it is not an integer.
        is_integer = isinstance(new_last_modified, int)
        if not is_integer:
            new.pop(modified_field, None)
            return new

        # Drop the new last_modified if lesser or equal to the old one.
        is_less_or_equal = (old is not None and
                            new_last_modified <= old[modified_field])
        if is_less_or_equal:
            new.pop(modified_field, None)

        return new

    def apply_changes(self, record, requested_changes):
        """Merge `changes` into `record` fields.

        .. note::

            This is used in the context of PATCH only.

        Override this to control field changes at record level, for example:

        .. code-block:: python

            def apply_changes(self, record, requested_changes):
                # Ignore value change if inferior
                if record['position'] > changes.get('position', -1):
                    changes.pop('position', None)
                return super().apply_changes(record, requested_changes)

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
            if result does not comply with resource schema.

        :returns: the new record with `changes` applied.
        :rtype: tuple
        """
        if self._is_json_patch:
            try:
                applied_changes = apply_json_patch(record, requested_changes)['data']
                updated = {**applied_changes}
            except ValueError as e:
                error_details = {
                    'location': 'body',
                    'description': 'JSON Patch operation failed: {}'.format(e)
                }
                raise_invalid(self.request, **error_details)

        else:
            applied_changes = {**requested_changes}
            updated = {**record}

            content_type = str(self.request.headers.get('Content-Type')).lower()
            # recursive patch and remove field if null attribute is passed (RFC 7396)
            if content_type == 'application/merge-patch+json':
                recursive_update_dict(updated, applied_changes, ignores=[None])
            else:
                updated.update(**applied_changes)

        for field, value in applied_changes.items():
            has_changed = record.get(field, value) != value
            if self.schema.is_readonly(field) and has_changed:
                error_details = {
                    'name': field,
                    'description': 'Cannot modify {}'.format(field)
                }
                raise_invalid(self.request, **error_details)

        try:
            validated = self.schema().deserialize(updated)
        except colander.Invalid as e:
            # Transform the errors we got from colander into Cornice errors.
            # We could not rely on Service schema because the record should be
            # validated only once the changes are applied
            for field, error in e.asdict().items():  # pragma: no branch
                raise_invalid(self.request, name=field, description=error)

        return validated, applied_changes

    def postprocess(self, result, action=ACTIONS.READ, old=None):
        body = {
            'data': result
        }

        parent_id = self.get_parent_id(self.request)
        self.request.notify_resource_event(parent_id=parent_id,
                                           timestamp=self.timestamp,
                                           data=result,
                                           action=action,
                                           old=old)

        return body

    #
    # Internals
    #

    def _get_record_or_404(self, record_id):
        """Retrieve record from storage and raise ``404 Not found`` if missing.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotFound` if
            the record is not found.
        """
        if self.context and self.context.current_record:
            # Set during authorization. Save a storage hit.
            return self.context.current_record

        try:
            return self.model.get_record(record_id)
        except storage_exceptions.RecordNotFoundError:
            details = {
                "id": record_id,
                "resource_name": self.request.current_resource_name
            }
            response = http_error(HTTPNotFound(), errno=ERRORS.INVALID_RESOURCE_ID,
                                  details=details)
            raise response

    def _add_timestamp_header(self, response, timestamp=None):
        """Add current timestamp in response headers, when request comes in.

        """
        if timestamp is None:
            timestamp = self.timestamp
        # Pyramid takes care of converting.
        response.last_modified = timestamp / 1000.0
        # Return timestamp as ETag.
        response.headers['ETag'] = '"{}"'.format(timestamp)

    def _add_cache_header(self, response):
        """Add Cache-Control and Expire headers, based a on a setting for the
        current resource.

        Cache headers will be set with anonymous requests only.

        .. note::

            The ``Cache-Control: no-cache`` response header does not prevent
            caching in client. It will indicate the client to revalidate
            the response content on each access. The client will send a
            conditional request to the server and check that a
            ``304 Not modified`` is returned before serving content from cache.
        """
        resource_name = self.context.resource_name if self.context else ''
        setting_key = '{}_cache_expires_seconds'.format(resource_name)
        collection_expires = self.request.registry.settings.get(setting_key)
        is_anonymous = self.request.prefixed_userid is None
        if collection_expires and is_anonymous:
            response.cache_expires(seconds=int(collection_expires))
        else:
            # Since `Expires` response header provides an HTTP data with a
            # resolution in seconds, do not use Pyramid `cache_expires()` in
            # order to omit it.
            response.cache_control.no_cache = True
            response.cache_control.no_store = True

    def _raise_400_if_invalid_id(self, record_id):
        """Raise 400 if specified record id does not match the format excepted
        by storage backends.

        :raises: :class:`pyramid.httpexceptions.HTTPBadRequest`
        """
        is_string = isinstance(record_id, str)
        if not is_string or not self.model.id_generator.match(record_id):
            error_details = {
                'location': 'path',
                'description': "Invalid record id"
            }
            raise_invalid(self.request, **error_details)

    def _raise_304_if_not_modified(self, record=None):
        """Raise 304 if current timestamp is inferior to the one specified
        in headers.

        :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPNotModified`
        """
        if_none_match = self.request.validated['header'].get('If-None-Match')

        if not if_none_match:
            return

        if if_none_match == '*':
            return

        if record:
            current_timestamp = record[self.model.modified_field]
        else:
            current_timestamp = self.model.timestamp()

        if current_timestamp == if_none_match:
            response = HTTPNotModified()
            self._add_timestamp_header(response, timestamp=current_timestamp)
            raise response

    def _raise_412_if_modified(self, record=None):
        """Raise 412 if current timestamp is superior to the one
        specified in headers.

        :raises:
            :exc:`~pyramid:pyramid.httpexceptions.HTTPPreconditionFailed`
        """
        if_match = self.request.validated['header'].get('If-Match')
        if_none_match = self.request.validated['header'].get('If-None-Match')

        # Check if record exists
        record_exists = record is not None

        # If no precondition headers, just ignore
        if not if_match and not if_none_match:
            return

        # If-None-Match: * should always raise if a record exists
        if if_none_match == '*' and record_exists:
            modified_since = -1  # Always raise.

        # If-Match should always raise if a record doesn't exist
        elif if_match and not record_exists:
            modified_since = -1

        # If-Match with ETag value on existing records should compare ETag
        elif if_match and if_match != '*':
            modified_since = if_match

        # If none of the above applies, don't raise
        else:
            return

        if record:
            current_timestamp = record[self.model.modified_field]
        else:
            current_timestamp = self.model.timestamp()

        if current_timestamp != modified_since:
            error_msg = 'Resource was modified meanwhile'
            details = {'existing': record} if record else {}
            response = http_error(HTTPPreconditionFailed(),
                                  errno=ERRORS.MODIFIED_MEANWHILE,
                                  message=error_msg,
                                  details=details)
            self._add_timestamp_header(response, timestamp=current_timestamp)
            raise response

    def _raise_400_if_id_mismatch(self, new_id, record_id):
        """Raise 400 if the `new_id`, within the request body, does not match
        the `record_id`, obtained from request path.

        :raises: :class:`pyramid.httpexceptions.HTTPBadRequest`
        """
        if new_id != record_id:
            error_msg = 'Record id does not match existing record'
            error_details = {
                'name': self.model.id_field,
                'description': error_msg
            }
            raise_invalid(self.request, **error_details)

    def _extract_partial_fields(self):
        """Extract the fields to do the projection from QueryString parameters.
        """
        fields = self.request.validated['querystring'].get('_fields')
        if fields:
            root_fields = [f.split('.')[0] for f in fields]
            known_fields = self._get_known_fields()
            invalid_fields = set(root_fields) - set(known_fields)
            preserve_unknown = self.schema.get_option('preserve_unknown')
            if not preserve_unknown and invalid_fields:
                error_msg = "Fields {} do not exist".format(','.join(invalid_fields))
                error_details = {
                    'name': "Invalid _fields parameter",
                    'description': error_msg
                }
                raise_invalid(self.request, **error_details)

            # Since id and last_modified are part of the synchronisation
            # API, force their presence in payloads.
            fields = fields + [self.model.id_field, self.model.modified_field]

        return fields

    def _extract_limit(self):
        """Extract limit value from QueryString parameters."""
        paginate_by = self.request.registry.settings['paginate_by']
        limit = self.request.validated['querystring'].get('_limit', paginate_by)

        # If limit is higher than paginate_by setting, ignore it.
        if limit and paginate_by:
            limit = min(limit, paginate_by)

        return limit

    def _extract_filters(self):
        """Extracts filters from QueryString parameters."""
        queryparams = self.request.validated['querystring']

        filters = []

        for param, value in queryparams.items():
            param = param.strip()

            error_details = {
                'name': param,
                'location': 'querystring',
                'description': 'Invalid value for {}'.format(param)
            }

            # Ignore specific fields
            if param.startswith('_') and param not in ('_since',
                                                       '_to',
                                                       '_before'):
                continue

            # Handle the _since specific filter.
            if param in ('_since', '_to', '_before'):

                if param == '_since':
                    operator = COMPARISON.GT
                else:
                    if param == '_to':
                        message = ('_to is now deprecated, '
                                   'you should use _before instead')
                        url = ('https://kinto.readthedocs.io/en/2.4.0/api/'
                               'resource.html#list-of-available-url-'
                               'parameters')
                        send_alert(self.request, message, url)
                    operator = COMPARISON.LT
                filters.append(
                    Filter(self.model.modified_field, value, operator)
                )
                continue

            allKeywords = '|'.join([i.name.lower() for i in COMPARISON])
            m = re.match(r'^('+allKeywords+')_([\w\.]+)$', param)
            if m:
                keyword, field = m.groups()
                operator = getattr(COMPARISON, keyword.upper())
            else:
                operator, field = COMPARISON.EQ, param

            if not self.is_known_field(field):
                error_msg = "Unknown filter field '{}'".format(param)
                error_details['description'] = error_msg
                raise_invalid(self.request, **error_details)

            if operator in (COMPARISON.IN, COMPARISON.EXCLUDE):
                all_integers = all([isinstance(v, int)
                                    for v in value])
                all_strings = all([isinstance(v, str)
                                   for v in value])
                has_invalid_value = (
                    (field == self.model.id_field and not all_strings) or
                    (field == self.model.modified_field and not all_integers)
                )
                if has_invalid_value:
                    raise_invalid(self.request, **error_details)

            filters.append(Filter(field, value, operator))

        return filters

    def _extract_sorting(self, limit):
        """Extracts filters from QueryString parameters."""
        specified = self.request.validated['querystring'].get('_sort', [])
        sorting = []
        modified_field_used = self.model.modified_field in specified
        for field in specified:
            field = field.strip()
            m = re.match(r'^([\-+]?)([\w\.]+)$', field)
            if m:
                order, field = m.groups()

                if not self.is_known_field(field):
                    error_details = {
                        'location': 'querystring',
                        'description': "Unknown sort field '{}'".format(field)
                    }
                    raise_invalid(self.request, **error_details)

                direction = -1 if order == '-' else 1
                sorting.append(Sort(field, direction))

        if not modified_field_used:
            # Add a sort by the ``modified_field`` in descending order
            # useful for pagination
            sorting.append(Sort(self.model.modified_field, -1))
        return sorting

    def _build_pagination_rules(self, sorting, last_record, rules=None):
        """Return the list of rules for a given sorting attribute and
        last_record.

        """
        if rules is None:
            rules = []

        rule = []
        next_sorting = sorting[:-1]

        for field, _ in next_sorting:
            rule.append(Filter(field, last_record.get(field), COMPARISON.EQ))

        field, direction = sorting[-1]

        if direction == -1:
            rule.append(Filter(field, last_record.get(field), COMPARISON.LT))
        else:
            rule.append(Filter(field, last_record.get(field), COMPARISON.GT))

        rules.append(rule)

        if len(next_sorting) == 0:
            return rules

        return self._build_pagination_rules(next_sorting, last_record, rules)

    def _extract_pagination_rules_from_token(self, limit, sorting):
        """Get pagination params."""
        token = self.request.validated['querystring'].get('_token', None)
        filters = []
        offset = 0
        if token:
            try:
                tokeninfo = json.loads(decode64(token))
                if not isinstance(tokeninfo, dict):
                    raise ValueError()
                last_record = tokeninfo['last_record']
                offset = tokeninfo['offset']
            except (ValueError, KeyError, TypeError):
                error_msg = '_token has invalid content'
                error_details = {
                    'location': 'querystring',
                    'description': error_msg
                }
                raise_invalid(self.request, **error_details)

            filters = self._build_pagination_rules(sorting, last_record)
        return filters, offset

    def _next_page_url(self, sorting, limit, last_record, offset):
        """Build the Next-Page header from where we stopped."""
        token = self._build_pagination_token(sorting, last_record, offset)

        params = {**request_GET(self.request), '_limit': limit, '_token': token}

        service = self.request.current_service
        next_page_url = self.request.route_url(service.name, _query=params,
                                               **self.request.matchdict)
        return next_page_url

    def _build_pagination_token(self, sorting, last_record, offset):
        """Build a pagination token.

        It is a base64 JSON object with the sorting fields values of
        the last_record.

        """
        token = {
            'last_record': {},
            'offset': offset
        }

        for field, _ in sorting:
            last_value = find_nested_value(last_record, field)
            if last_value is not None:
                token['last_record'][field] = last_value

        return encode64(json.dumps(token))


class ShareableResource(UserResource):
    """Shareable resources allow to set permissions on records, in order to
    share their access or protect their modification.
    """
    default_model = ShareableModel
    default_viewset = ShareableViewSet
    permissions = ('read', 'write')
    """List of allowed permissions names."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # In base resource, PATCH only hit storage if no data has changed.
        # Here, we force update because we add the current principal to
        # the ``write`` ACE.
        self.force_patch_update = True

        # Required by the ShareableModel class.
        self.model.permission = self.request.registry.permission
        if self.request.prefixed_userid is None:
            # The principal of an anonymous is system.Everyone
            self.model.current_principal = Everyone
        else:
            self.model.current_principal = self.request.prefixed_userid
        self.model.prefixed_principals = self.request.prefixed_principals

        if self.context:
            self.model.get_permission_object_id = functools.partial(
                self.context.get_permission_object_id,
                self.request)

    def get_parent_id(self, request):
        """Unlike :class:`kinto.core.resource.UserResource`, records are not
        isolated by user.

        See https://github.com/mozilla-services/cliquet/issues/549

        :returns: A constant empty value.
        """
        return ''

    def _extract_filters(self):
        """Override default filters extraction from QueryString to allow
        partial collection of records.

        XXX: find more elegant approach to add custom filters.
        """
        filters = super()._extract_filters()

        ids = self.context.shared_ids
        if ids is not None:
            filter_by_id = Filter(self.model.id_field, ids, COMPARISON.IN)
            filters.insert(0, filter_by_id)

        return filters

    def _raise_412_if_modified(self, record=None):
        """Do not provide the permissions among the record fields.
        Ref: https://github.com/Kinto/kinto/issues/224
        """
        if record:
            record = {**record}
            record.pop(self.model.permissions_field, None)
        return super()._raise_412_if_modified(record)

    def process_record(self, new, old=None):
        """Read permissions from request body, and in the case of ``PUT`` every
        existing ACE is removed (using empty list).
        """
        new = super().process_record(new, old)

        # patch is specified as a list of of operations (RFC 6902)
        if self._is_json_patch:
            changes = self.request.validated['body']
            permissions = apply_json_patch(old, changes)['permissions']
        else:
            permissions = self.request.validated['body'].get('permissions', {})

        annotated = {**new}

        if permissions:
            is_put = (self.request.method.lower() == 'put')
            if is_put:
                # Remove every existing ACEs using empty lists.
                for perm in self.permissions:
                    permissions.setdefault(perm, [])
            annotated[self.model.permissions_field] = permissions

        return annotated

    def postprocess(self, result, action=ACTIONS.READ, old=None):
        """Add ``permissions`` attribute in response body.

        In the HTTP API, it was decided that ``permissions`` would reside
        outside the ``data`` attribute.
        """
        body = {}

        if not isinstance(result, list):
            # record endpoint.
            perms = result.pop(self.model.permissions_field, None)
            if perms is not None:
                body['permissions'] = {k: list(p) for k, p in perms.items()}

            if old:
                # Remove permissions from event payload.
                old.pop(self.model.permissions_field, None)

        data = super().postprocess(result, action, old)
        body.update(data)
        return body



import json
import os

import colander
from pyramid import httpexceptions
from pyramid.security import NO_PERMISSION_REQUIRED
from kinto.core import Service

HERE = os.path.dirname(os.path.abspath(__file__))
ORIGIN = os.path.dirname(os.path.dirname(HERE))


class VersionResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown='preserve'))


version_response_schemas = {
    '200': VersionResponseSchema(
        description='Return the running Instance version information.')
}


version = Service(name="version", path='/__version__', description="Version")


@version.get(permission=NO_PERMISSION_REQUIRED, tags=['Utilities'],
             operation_id='__version__', response_schemas=version_response_schemas)
def version_view(request):
    try:
        return version_view.__json__
    except AttributeError:
        pass

    location = request.registry.settings['version_json_path']
    files = [
        location,  # Default is current working dir.
        os.path.join(ORIGIN, 'version.json'),  # Relative to the package root.
        os.path.join(HERE, 'version.json')  # Relative to this file.
    ]
    for version_file in files:
        file_path = os.path.abspath(version_file)
        if os.path.exists(file_path):
            with open(file_path) as f:
                version_view.__json__ = json.load(f)
                return version_view.__json__  # First one wins.

    raise httpexceptions.HTTPNotFound()



from kinto.authorization import PERMISSIONS_INHERITANCE_TREE
from kinto.core.events import ResourceChanged

from .listener import on_resource_changed


def includeme(config):
    config.add_api_capability(
        'history',
        description='Track changes on data.',
        url='http://kinto.readthedocs.io/en/latest/api/1.x/history.html')

    # Activate end-points.
    config.scan('kinto.plugins.history.views')

    # If StatsD is enabled, monitor execution time of listener.
    listener = on_resource_changed
    if config.registry.statsd:
        key = 'plugins.history'
        listener = config.registry.statsd.timer(key)(on_resource_changed)

    # Listen to every resources (except history)
    config.add_subscriber(listener, ResourceChanged,
                          for_resources=('bucket', 'group',
                                         'collection', 'record'))

    # Register the permission inheritance for history entries.
    PERMISSIONS_INHERITANCE_TREE['history'] = {
        'read': {
            'bucket': ['write', 'read'],
            'history': ['write', 'read']
        },
        'write': {
            'bucket': ['write'],
            'history': ['write']
        },
    }



import json
import re
import unittest
import mock

from pyramid import testing

from kinto import main as kinto_main
from kinto.core.testing import get_user_headers, skip_if_no_statsd

from .. import support


class PluginSetup(unittest.TestCase):

    @skip_if_no_statsd
    def test_a_statsd_timer_is_used_for_history_if_configured(self):
        settings = {
            "statsd_url": "udp://127.0.0.1:8125",
            "includes": "kinto.plugins.history"
        }
        config = testing.setUp(settings=settings)
        with mock.patch('kinto.core.statsd.Client.timer') as mocked:
            kinto_main(None, config=config)
            mocked.assert_called_with('plugins.history')


class HistoryWebTest(support.BaseWebTest, unittest.TestCase):

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['includes'] = 'kinto.plugins.history'
        return settings


class HelloViewTest(HistoryWebTest):

    def test_history_capability_if_enabled(self):
        resp = self.app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('history', capabilities)


class HistoryViewTest(HistoryWebTest):

    def setUp(self):
        self.bucket_uri = '/buckets/test'
        self.app.put(self.bucket_uri, headers=self.headers)

        self.collection_uri = self.bucket_uri + '/collections/col'
        resp = self.app.put(self.collection_uri, headers=self.headers)
        self.collection = resp.json['data']

        self.group_uri = self.bucket_uri + '/groups/grp'
        body = {'data': {'members': ['elle']}}
        resp = self.app.put_json(self.group_uri, body, headers=self.headers)
        self.group = resp.json['data']

        self.record_uri = '/buckets/test/collections/col/records/rec'
        body = {'data': {'foo': 42}}
        resp = self.app.put_json(self.record_uri, body, headers=self.headers)
        self.record = resp.json['data']

        self.history_uri = '/buckets/test/history'

    def test_only_get_and_delete_on_collection_are_allowed(self):
        self.app.put(self.history_uri, headers=self.headers, status=405)
        self.app.patch(self.history_uri, headers=self.headers, status=405)

    def test_only_collection_endpoint_is_available(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        url = '{}/{}'.format(self.bucket_uri, entry['id'])
        self.app.get(url, headers=self.headers, status=404)
        self.app.put(url, headers=self.headers, status=404)
        self.app.patch(url, headers=self.headers, status=404)
        self.app.delete(url, headers=self.headers, status=404)

    def test_tracks_user_and_date(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][-1]
        assert entry['user_id'] == self.principal
        assert re.match('^\d{4}\-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}',
                        entry['date'])

    #
    # Bucket
    #

    def test_history_contains_bucket_creation(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][-1]
        assert entry['resource_name'] == 'bucket'
        assert entry['bucket_id'] == 'test'
        assert entry['action'] == 'create'
        assert entry['uri'] == '/buckets/test'

    def test_history_supports_creation_via_plural_endpoint(self):
        resp = self.app.post_json('/buckets', {'data': {'id': 'posted'}},
                                  headers=self.headers)
        resp = self.app.get('/buckets/posted/history', headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['resource_name'] == 'bucket'
        assert entry['bucket_id'] == 'posted'
        assert entry['action'] == 'create'
        assert entry['uri'] == '/buckets/posted'

    def test_tracks_bucket_attributes_update(self):
        body = {'data': {'foo': 'baz'}}
        self.app.patch_json(self.bucket_uri, body,
                            headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'update'
        assert entry['target']['data']['foo'] == 'baz'

    def test_tracks_bucket_permissions_update(self):
        body = {'permissions': {'read': ['admins']}}
        self.app.patch_json(self.bucket_uri, body,
                            headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'update'
        assert entry['target']['permissions']['read'] == ['admins']

    def test_bucket_delete_destroys_its_history_entries(self):
        self.app.delete(self.bucket_uri, headers=self.headers)
        storage = self.app.app.registry.storage
        stored_in_backend, _ = storage.get_all(parent_id='/buckets/test',
                                               collection_id='history')
        assert len(stored_in_backend) == 0

    def test_delete_all_buckets_destroys_history_entries(self):
        self.app.put_json('/buckets/1', {"data": {"a": 1}},
                          headers=self.headers)

        self.app.delete('/buckets?a=1', headers=self.headers)

        # Entries about deleted bucket are gone.
        storage = self.app.app.registry.storage
        stored_in_backend, _ = storage.get_all(parent_id='/buckets/1',
                                               collection_id='history')
        assert len(stored_in_backend) == 0

        # Entries of other buckets are still here.
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][-1]
        assert entry['bucket_id'] == 'test'
        assert entry['action'] == 'create'

    #
    # Collection
    #

    def test_tracks_collection_creation(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][2]
        cid = self.collection['id']
        assert entry['resource_name'] == 'collection'
        assert 'bucket_id' not in entry
        assert entry['collection_id'] == cid
        assert entry['action'] == 'create'
        assert entry['uri'] == '/buckets/test/collections/{}'.format(cid)

    def test_tracks_collection_attributes_update(self):
        body = {'data': {'foo': 'baz'}}
        self.app.patch_json(self.collection_uri, body,
                            headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'update'
        assert entry['target']['data']['foo'] == 'baz'

    def test_tracks_collection_permissions_update(self):
        body = {'permissions': {'read': ['admins']}}
        self.app.patch_json(self.collection_uri, body,
                            headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'update'
        assert entry['target']['permissions']['read'] == ['admins']

    def test_tracks_collection_delete(self):
        self.app.delete(self.collection_uri, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'delete'
        assert entry['target']['data']['deleted'] is True

    def test_tracks_multiple_collections_delete(self):
        self.app.put(self.bucket_uri + '/collections/col2',
                     headers=self.headers)

        self.app.delete(self.bucket_uri + '/collections', headers=self.headers)

        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'delete'
        assert entry['target']['data']['id'] in (self.collection['id'], 'col2')

    #
    # Group
    #

    def test_tracks_group_creation(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][1]
        assert entry['resource_name'] == 'group'
        assert 'bucket_id' not in entry
        assert entry['group_id'] == self.group['id']
        assert entry['action'] == 'create'
        assert entry['uri'] == '/buckets/test/groups/{}'.format(self.group['id'])

    def test_tracks_group_attributes_update(self):
        body = {'data': {'foo': 'baz', 'members': ['lui']}}
        self.app.patch_json(self.group_uri, body,
                            headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'update'
        assert entry['target']['data']['foo'] == 'baz'
        assert entry['target']['data']['members'] == ['lui']

    def test_tracks_group_permissions_update(self):
        body = {'permissions': {'read': ['admins']}}
        self.app.patch_json(self.group_uri, body,
                            headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'update'
        assert entry['target']['permissions']['read'] == ['admins']

    def test_tracks_group_delete(self):
        self.app.delete(self.group_uri, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'delete'
        assert entry['target']['data']['deleted'] is True

    def test_tracks_multiple_groups_delete(self):
        self.app.put_json(self.bucket_uri + '/groups/g2',
                          {"data": {"members": ["her"]}},
                          headers=self.headers)

        self.app.delete(self.bucket_uri + '/groups', headers=self.headers)

        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'delete'
        assert entry['target']['data']['id'] in (self.group['id'], 'g2')

    #
    # Record
    #

    def test_tracks_record_creation(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        cid = self.collection['id']
        rid = self.record['id']
        assert entry['resource_name'] == 'record'
        assert 'bucket_id' not in entry
        assert entry['collection_id'] == cid
        assert entry['record_id'] == rid
        assert entry['action'] == 'create'
        assert entry['uri'] == '/buckets/test/collections/{}/records/{}'.format(cid, rid)
        assert entry['target']['data']['foo'] == 42
        assert entry['target']['permissions']['write'][0].startswith('basicauth:')

    def test_tracks_record_attributes_update(self):
        resp = self.app.patch_json(self.record_uri, {'data': {'foo': 'baz'}},
                                   headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'update'
        assert entry['target']['data']['foo'] == 'baz'

    def test_tracks_record_permissions_update(self):
        body = {'permissions': {'read': ['admins']}}
        resp = self.app.patch_json(self.record_uri, body,
                                   headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'update'
        assert entry['target']['permissions']['read'] == ['admins']

    def test_tracks_record_delete(self):
        resp = self.app.delete(self.record_uri, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'delete'
        assert entry['target']['data']['deleted'] is True

    def test_tracks_multiple_records_delete(self):
        records_uri = self.collection_uri + '/records'
        body = {'data': {'foo': 43}}
        resp = self.app.post_json(records_uri, body, headers=self.headers)
        rid = resp.json['data']['id']

        self.app.delete(records_uri, headers=self.headers)

        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json['data'][0]
        assert entry['action'] == 'delete'
        assert entry['target']['data']['id'] in (self.record['id'], rid)

    def test_does_not_track_records_during_massive_deletion(self):
        body = {'data': {'pim': 'pam'}}
        records_uri = self.collection_uri + '/records'
        self.app.post_json(records_uri, body, headers=self.headers)

        self.app.delete(self.collection_uri, headers=self.headers)

        resp = self.app.get(self.history_uri, headers=self.headers)
        deletion_entries = [e for e in resp.json['data'] if e['action'] == 'delete']
        assert len(deletion_entries) == 1


class HistoryDeletionTest(HistoryWebTest):

    def setUp(self):
        self.app.put('/buckets/bid', headers=self.headers)
        self.app.put('/buckets/bid/collections/cid',
                     headers=self.headers)
        body = {'data': {'foo': 42}}
        self.app.put_json('/buckets/bid/collections/cid/records/rid',
                          body,
                          headers=self.headers)

    def test_full_deletion(self):
        self.app.delete('/buckets/bid/history', headers=self.headers)
        resp = self.app.get('/buckets/bid/history', headers=self.headers)
        assert len(resp.json['data']) == 0

    def test_partial_deletion(self):
        resp = self.app.get('/buckets/bid/history', headers=self.headers)
        before = int(json.loads(resp.headers['ETag']))
        self.app.put('/buckets/bid/collections/cid2', headers=self.headers)

        # Delete everything before the last entry (exclusive)
        self.app.delete('/buckets/bid/history?_before={}'.format(before),
                        headers=self.headers)

        resp = self.app.get('/buckets/bid/history', headers=self.headers)
        assert len(resp.json['data']) == 2  # record + new collection


class FilteringTest(HistoryWebTest):

    def setUp(self):
        self.app.put('/buckets/bid', headers=self.headers)
        self.app.put('/buckets/0', headers=self.headers)
        self.app.put('/buckets/bid/collections/cid',
                     headers=self.headers)
        self.app.put('/buckets/0/collections/1',
                     headers=self.headers)
        body = {'data': {'foo': 42}}
        self.app.put_json('/buckets/bid/collections/cid/records/rid',
                          body,
                          headers=self.headers)
        body = {'data': {'foo': 0}}
        self.app.put_json('/buckets/0/collections/1/records/2',
                          body,
                          headers=self.headers)
        body = {'data': {'foo': 'bar'}}
        self.app.patch_json('/buckets/bid/collections/cid/records/rid',
                            body,
                            headers=self.headers)
        self.app.delete('/buckets/bid/collections/cid/records/rid',
                        headers=self.headers)

    def test_filter_by_unknown_field_is_not_allowed(self):
        self.app.get('/buckets/bid/history?movie=bourne',
                     headers=self.headers,
                     status=400)

    def test_filter_by_action(self):
        resp = self.app.get('/buckets/bid/history?action=delete',
                            headers=self.headers)
        assert len(resp.json['data']) == 1

    def test_filter_by_resource(self):
        resp = self.app.get('/buckets/bid/history?resource_name=bucket',
                            headers=self.headers)
        assert len(resp.json['data']) == 1

    def test_filter_by_uri(self):
        uri = '/buckets/bid/collections/cid/records/rid'
        resp = self.app.get('/buckets/bid/history?uri={}'.format(uri),
                            headers=self.headers)
        assert len(resp.json['data']) == 3  # create / update / delete

    def test_allows_diff_between_two_versions_of_a_record(self):
        uri = '/buckets/bid/collections/cid/records/rid'
        querystring = '?uri={}&_limit=2&_sort=last_modified'.format(uri)
        resp = self.app.get('/buckets/bid/history{}'.format(querystring),
                            headers=self.headers)
        entries = resp.json['data']
        version1 = entries[0]['target']['data']
        version2 = entries[1]['target']['data']
        diff = [(k, version1[k], version2[k])
                for k in version1.keys()
                if k != 'last_modified' and version2[k] != version1[k]]
        assert diff == [('foo', 42, 'bar')]

    def test_filter_by_bucket(self):
        uri = '/buckets/bid/history?bucket_id=bid'
        resp = self.app.get(uri,
                            headers=self.headers)
        # This is equivalent to filtering by resource_name=bucket,
        # since only entries for bucket have ``bucket_id`` attribute.
        assert len(resp.json['data']) == 1

    def test_filter_by_collection(self):
        uri = '/buckets/bid/history?collection_id=cid'
        resp = self.app.get(uri,
                            headers=self.headers)
        assert len(resp.json['data']) == 4

    def test_filter_by_numeric_bucket(self):
        uri = '/buckets/0/history?bucket_id=0'
        resp = self.app.get(uri,
                            headers=self.headers)
        assert len(resp.json['data']) == 1

    def test_filter_by_numeric_collection(self):
        uri = '/buckets/0/history?collection_id=1'
        resp = self.app.get(uri,
                            headers=self.headers)
        assert len(resp.json['data']) == 2

    def test_filter_by_numeric_record(self):
        uri = '/buckets/0/history?record_id=2'
        resp = self.app.get(uri,
                            headers=self.headers)
        assert len(resp.json['data']) == 1

    def test_filter_by_target_fields(self):
        uri = '/buckets/bid/history?target.data.id=rid'
        resp = self.app.get(uri,
                            headers=self.headers)
        assert len(resp.json['data']) == 3  # create, update, delete

    def test_limit_results(self):
        resp = self.app.get('/buckets/bid/history?_limit=2',
                            headers=self.headers)
        assert len(resp.json['data']) == 2
        assert 'Next-Page' in resp.headers

    def test_filter_returned_fields(self):
        resp = self.app.get('/buckets/bid/history?_fields=uri,action',
                            headers=self.headers)
        assert sorted(resp.json['data'][0].keys()) == ['action', 'id',
                                                       'last_modified', 'uri']

    def test_sort_by_date(self):
        resp = self.app.get('/buckets/bid/history?_sort=date',
                            headers=self.headers)
        entries = resp.json['data']
        assert entries[0]['date'] < entries[-1]['date']


class BulkTest(HistoryWebTest):
    def setUp(self):
        body = {
            'defaults': {
                'method': 'POST',
                'path': '/buckets/bid/collections/cid/records',
            },
            'requests': [{
                'path': '/buckets/bid',
                'method': 'PUT'
            }, {
                'path': '/buckets/bid/collections',
                'body': {'data': {'id': 'cid'}}
            }, {
                'body': {'data': {'id': 'a', 'attr': 1}},
            }, {
                'body': {'data': {'id': 'b', 'attr': 2}},
            }, {
                'body': {'data': {'id': 'c', 'attr': 3}}
            }]
        }
        self.app.post_json('/batch', body, headers=self.headers)

    def test_post_on_collection(self):
        resp = self.app.get('/buckets/bid/history', headers=self.headers)
        entries = resp.json['data']
        assert len(entries) == 5
        assert entries[0]['uri'] == '/buckets/bid/collections/cid/records/c'
        assert entries[-2]['uri'] == '/buckets/bid/collections/cid'

    def test_delete_on_collection(self):
        body = {
            'defaults': {
                'method': 'DELETE',
            },
            'requests': [{
                'path': '/buckets/bid/collections/cid/records/a',
            }, {
                'path': '/buckets/bid/collections/cid/records/b',
            }, {
                'path': '/buckets/bid/collections/cid/records/c',
            }]
        }
        self.app.post_json('/batch', body, headers=self.headers)
        resp = self.app.get('/buckets/bid/history', headers=self.headers)
        entries = resp.json['data']
        assert entries[0]['uri'] == '/buckets/bid/collections/cid/records/c'
        assert entries[1]['uri'] == '/buckets/bid/collections/cid/records/b'
        assert entries[2]['uri'] == '/buckets/bid/collections/cid/records/a'

    def test_multiple_patch(self):
        # Kinto/kinto#942
        requests = [{
            'method': 'PATCH',
            'path': '/buckets/bid/collections/cid/records/{}'.format(l),
            'body': {'data': {'label': l}}} for l in ('a', 'b', 'c')]
        self.app.post_json('/batch', {'requests': requests}, headers=self.headers)
        resp = self.app.get('/buckets/bid/history', headers=self.headers)
        entries = resp.json['data']
        for entry in entries:
            if entry['resource_name'] != 'record':
                continue
            assert entry['record_id'] == entry['target']['data']['id']


class DefaultBucketTest(HistoryWebTest):

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['includes'] = ('kinto.plugins.default_bucket '
                                'kinto.plugins.history')
        return settings

    def setUp(self):
        resp = self.app.get('/', headers=self.headers)
        self.bucket_id = resp.json['user']['bucket']
        self.history_uri = '/buckets/{}/history'.format(self.bucket_id)

    def test_history_can_be_accessed_via_default_alias(self):
        self.app.get('/buckets/default/collections/blah',
                     headers=self.headers)
        resp = self.app.get('/buckets/default/history', headers=self.headers)
        assert len(resp.json['data']) == 2

    def test_implicit_creations_are_listed(self):
        body = {'data': {'foo': 42}}
        resp = self.app.post_json('/buckets/default/collections/blah/records',
                                  body,
                                  headers=self.headers)
        record = resp.json['data']

        resp = self.app.get(self.history_uri, headers=self.headers)
        entries = resp.json['data']
        assert len(entries) == 3

        bucket_uri = '/buckets/{}'.format(self.bucket_id)
        assert entries[2]['resource_name'] == 'bucket'
        assert entries[2]['bucket_id'] == self.bucket_id
        assert entries[2]['uri'] == bucket_uri
        assert entries[2]['target']['permissions']['write'][0] == self.principal

        collection_uri = bucket_uri + '/collections/blah'
        assert entries[1]['resource_name'] == 'collection'
        assert 'bucket_id' not in entries[1]
        assert entries[1]['collection_id'] == 'blah'
        assert entries[1]['uri'] == collection_uri
        assert entries[1]['target']['permissions']['write'][0] == self.principal

        record_uri = collection_uri + '/records/{}'.format(record['id'])
        assert entries[0]['resource_name'] == 'record'
        assert 'bucket_id' not in entries[1]
        assert entries[0]['collection_id'] == 'blah'
        assert entries[0]['record_id'] == record['id']
        assert entries[0]['uri'] == record_uri
        assert entries[0]['target']['data']['foo'] == 42
        assert entries[0]['target']['permissions']['write'][0] == self.principal


class PermissionsTest(HistoryWebTest):

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['experimental_permissions_endpoint'] = 'true'
        return settings

    def setUp(self):
        self.alice_headers = get_user_headers('alice')
        self.bob_headers = get_user_headers('bob')
        self.julia_headers = get_user_headers('julia')

        self.alice_principal = ('basicauth:d5b0026601f1b251974e09548d44155e16'
                                '812e3c64ff7ae053fe3542e2ca1570')
        self.bob_principal = ('basicauth:c031ced27503f788b102ca54269a062ec737'
                              '94bb075154c74a0d4311e74ca8b6')
        self.julia_principal = ('basicauth:d8bab8d9fe0510fcaf9b5ad5942c027fc'
                                '2fdf80b6dc59cc3c48d12a2fcb18f1c')

        bucket = {
            'permissions': {
                'read': [self.alice_principal]
            }
        }
        collection = {
            'permissions': {
                'read': [self.julia_principal]
            }
        }
        record = {
            'permissions': {
                'write': [self.bob_principal, self.alice_principal],
            }
        }
        self.app.put('/buckets/author-only',
                     headers=self.headers)
        self.app.put_json('/buckets/test',
                          bucket,
                          headers=self.headers)
        self.app.put_json('/buckets/test/groups/admins',
                          {'data': {'members': []}},
                          headers=self.headers)
        self.app.put_json('/buckets/test/collections/alice-julia',
                          collection,
                          headers=self.headers)
        self.app.put_json('/buckets/test/collections/author-only',
                          headers=self.headers)
        self.app.post_json('/buckets/test/collections/alice-julia/records',
                           record,
                           headers=self.headers)
        self.app.post_json('/buckets/test/collections/alice-julia/records',
                           {'permissions': {'read': ['system.Authenticated']}},
                           headers=self.headers)

    def test_author_can_read_everything(self):
        resp = self.app.get('/buckets/test/history',
                            headers=self.headers)
        entries = resp.json['data']
        assert len(entries) == 6  # everything.

    def test_read_permission_can_be_given_to_anybody_via_settings(self):
        with mock.patch.dict(self.app.app.registry.settings,
                             [('history_read_principals', 'system.Everyone')]):
            resp = self.app.get('/buckets/test/history',
                                headers=get_user_headers('tartan:pion'))
            entries = resp.json['data']
            assert len(entries) == 6  # everything.

    def test_bucket_read_allows_whole_history(self):
        resp = self.app.get('/buckets/test/history',
                            headers=self.alice_headers)
        entries = resp.json['data']
        assert len(entries) == 6  # everything.

        self.app.get('/buckets/author-only/history',
                     headers=self.alice_headers,
                     status=403)

    def test_collection_read_restricts_to_collection(self):
        resp = self.app.get('/buckets/test/history',
                            headers=self.julia_headers)
        entries = resp.json['data']
        assert len(entries) == 3
        assert entries[0]['resource_name'] == 'record'
        assert entries[1]['resource_name'] == 'record'
        assert entries[2]['resource_name'] == 'collection'

    def test_write_on_record_restricts_to_record(self):
        resp = self.app.get('/buckets/test/history',
                            headers=self.bob_headers)
        entries = resp.json['data']
        assert len(entries) == 2
        assert ('system.Authenticated' in
                entries[0]['target']['permissions']['read'])
        assert entries[0]['resource_name'] == 'record'
        assert (self.bob_principal in
                entries[1]['target']['permissions']['write'])
        assert entries[1]['resource_name'] == 'record'

    def test_publicly_readable_record_allows_any_authenticated(self):
        resp = self.app.get('/buckets/test/history',
                            headers=get_user_headers('jack:'))
        entries = resp.json['data']
        assert len(entries) == 1
        assert ('system.Authenticated' in
                entries[0]['target']['permissions']['read'])
        assert entries[0]['resource_name'] == 'record'

    def test_new_entries_are_not_readable_if_permission_is_removed(self):
        resp = self.app.get('/buckets/test/history',
                            headers=self.alice_headers)
        before = resp.headers['ETag']

        # Remove alice from read permission.
        self.app.patch_json('/buckets/test',
                            {'permissions': {'read': []}},
                            headers=self.headers)

        # Create new collection.
        self.app.put_json('/buckets/test/collections/new-one',
                          headers=self.headers)

        # History did not evolve for alice.
        resp = self.app.get('/buckets/test/history',
                            headers=self.alice_headers)
        assert resp.headers['ETag'] != before

    def test_history_entries_are_not_listed_in_permissions_endpoint(self):
        resp = self.app.get('/permissions',
                            headers=self.headers)
        entries = [e['resource_name'] == 'history' for e in resp.json["data"]]
        assert not any(entries)


class ExcludeResourcesTest(HistoryWebTest):

    def get_app_settings(self, extras=None):
        settings = super().get_app_settings(extras)
        settings['history.exclude_resources'] = ('/buckets/a '
                                                 '/buckets/b/collections/a '
                                                 '/buckets/b/groups/a')
        return settings

    def setUp(self):
        group = {'data': {'members': []}}
        self.app.put_json('/buckets/a', headers=self.headers)
        self.app.put_json('/buckets/a/groups/admins', group, headers=self.headers)
        self.app.put_json('/buckets/b', headers=self.headers)
        self.app.put_json('/buckets/b/groups/a', group, headers=self.headers)
        self.app.put_json('/buckets/b/collections/a', headers=self.headers)
        self.app.put_json('/buckets/b/collections/a/records/1', headers=self.headers)
        self.app.put_json('/buckets/b/collections/b', headers=self.headers)
        self.app.put_json('/buckets/b/collections/b/records/1', headers=self.headers)

    def test_whole_buckets_can_be_excluded(self):
        resp = self.app.get('/buckets/a/history',
                            headers=self.headers)
        entries = resp.json['data']
        assert len(entries) == 0  # nothing.

    def test_some_specific_collection_can_be_excluded(self):
        resp = self.app.get('/buckets/b/history?collection_id=b',
                            headers=self.headers)
        entries = resp.json['data']
        assert len(entries) > 0

        resp = self.app.get('/buckets/b/history?collection_id=a',
                            headers=self.headers)
        entries = resp.json['data']
        assert len(entries) == 0  # nothing.

    def test_some_specific_object_can_be_excluded(self):
        resp = self.app.get('/buckets/b/history?group_id=a',
                            headers=self.headers)
        entries = resp.json['data']
        assert len(entries) == 0  # nothing.
